# Copyright 2020-present, Mayo Clinic Department of Neurology - Bioelectronics Neurophysiology and Engineering Laboratory
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from best.dbs.artifact_bank import stimulation_artifacts
from best.dbs._generator import ArtifactGenerator
from best.dbs.artifact_removal.dataset import StimArtifactDataset
from best.dbs.artifact_removal.model import dbs_artifact_removal_network


















# Copyright 2020-present, Mayo Clinic Department of Neurology - Bioelectronics Neurophysiology and Engineering Laboratory
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import os
import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod

import os
# Check windows or linux and sets separator
if os.name == 'nt': DELIMITER = '\\'
else: DELIMITER = '/'



class myXML(ABC):
    """
    General function for XML Parsing utilizing xml.etree.ElementTree library and Elements of this lib.
    This function represents a single tag in XML structure.
    E.g.
    <meta attr1='0'>
    <time> '01-02-03' <\time>
    <\meta>
    etc.
    Object meta is defined by its tag 'meta' and and attribute 0. Has also the child with tag 'time'.
    Sub-object 'time' has only param '01-02-03'

    Object myXML is never meant to be called. Instead it is template for creating new objects inheriting from this instance.
    See bellow. Proper class definition is is:

    class myXML_Segment(myXML):
    tag = 'segment'
    ref_children = [myXML_url, myXML_Stop, myXML_Start, myXML_Id]

    tag is string exactly how it occurs in the XML file. ref_children is a list of objects which can be as sub-objects.
    This way, a proper structure can be defined and its integrity is ensured.
    If different tag appears than defined error is raised and structure must be defined.

    """

    tag = ''
    ref_children = []
    def __init__(self, Parent):
        """
        Init func which is NEVER used. Always use methods init_EmptyInstance or init_from_existing to initialize object myXML.
        """
        self._version = Parent._version
        self._file_id = Parent._file_id
        self._file_path = Parent._file_path

        self._children = {} # initialized in the init_from_existing func
        for ref_child in self.ref_children:
            self._children[ref_child.tag] = []

        for key in self._children.keys():
            setattr(self, key, self._children[key])

        self.attributes = myXML_AttributeHandler

    @classmethod
    def init_from_existing(cls, Parent, Element: ET.Element):
        """
        Main function to initialize instance from existing ET.Element which is read from the XML File.
        Inherits properties from Parent instance see __init__(). Element is a representation of a Tag using xml.etree lib.

        """
        iniClass = cls(Parent)
        # assert Element, 'Tag ' + iniClass.tag + 'is not declared in the processed file ' + Parent.file_path
        if not iniClass._set_Element(Element):
            raise AttributeError('Tag \'' + iniClass.tag + '\' is not declared in the processed file ' + Parent._file_path)

        iniClass.attributes = myXML_AttributeHandler(Element)
        iniClass._init_children()

        return iniClass

    @classmethod
    def init_EmptyInstance(cls, file_path = '', Parent = None, param = '', attributes = None):
        """
        Initialize empty instance

        """
        file_id = file_path.split('.')[-2].split(DELIMITER)[-1]

        if not Parent:
            class temp_class:
                def __init__(self, version='', file_id='', file_path=''):
                    self._version = version
                    self._file_id = file_id
                    self._file_path = file_path
                    self.tag = cls.tag
                    self.Element = ET.Element(cls.tag)

        Parent = temp_class(version="1.0", file_id=file_id, file_path=file_path)
        NewInstance = cls(Parent)
        NewInstance._set_Element(ET.Element(cls.tag))
        NewInstance.attributes = myXML_AttributeHandler(NewInstance.Element)
        NewInstance.param = param
        return NewInstance

    def add_child(self, child_initializer, param="", attributes={}):
        child = child_initializer(self)
        child.Element = ET.SubElement(self.Element, child.tag)
        child.attributes = myXML_AttributeHandler(child.Element)

        child.param = param
        for key in attributes.keys():
            child.attributes[key] = attributes[key]

        is_child_set = False
        for ref_child_key in self._children.keys():
            if ref_child_key == child.tag:
                self._children[child.tag].append(child)
                is_child_set = True

        if is_child_set is False:
            raise KeyError('Children \"' + child.tag + '\" cannot be assigned to a parent \"' + self.tag + '\". Defined children are: ' + str(self._children.keys()))

        return child

    def _init_children(self):
        """
        Initialize all children in a loop.
        """

        def _init_children_struct(ref_children):
            _children = {}
            for ref_child in ref_children:
                _children[ref_child.tag] = []
            return _children

        self._children = _init_children_struct(self.ref_children)
        for childEl in self.Element:
            self._init_child(childEl)

    def _init_child(self, ChildElement):
        """
        Inits a child from element

        """
        childInitializator = self._find_childClass(ChildElement.tag.split('}')[-1])
        assert childInitializator, 'Tag: ' + ChildElement.tag + \
                                              ' is not correctly declared in the structure' \
                                              ' of XML file. Please, have a look into the myXML' \
                                              'and add appropriate class and references into this package.'

        if type(self._children[ChildElement.tag.split('}')[-1]]) is list and self._children[ChildElement.tag.split('}')[-1]].__len__() == 0:
            self._children[ChildElement.tag.split('}')[-1]] = childInitializator.init_from_existing(self, ChildElement)
        elif not (type(self._children[ChildElement.tag.split('}')[-1]]) is list):
            self._children[ChildElement.tag.split('}')[-1]] = [self._children[ChildElement.tag.split('}')[-1]],
                                                childInitializator.init_from_existing(self, ChildElement)
                                                ]
        else:
            self._children[ChildElement.tag.split('}')[-1]].append(
                childInitializator.init_from_existing(self, ChildElement)
            )

    def _find_childClass(self, tag):
        for childClass in self.ref_children:
            if childClass.tag == tag:
                return childClass
        return False

    def _set_Element(self, aElement: ET.Element):
        if self._verify_input(aElement):
            self.Element = aElement
            return True
        return False

    def _verify_input(self, Element: ET.Element):
        ETag = Element.tag
        ETag = ETag.split('}')[-1]
        return ETag == self.tag

    def __getitem__(self, key):
        assert type(key) is str, 'Key to access child tags must be string!'
        assert key in self.keys(), 'Key \'' + '\' is not present in the existing tags.'
        return self._children[key]

    def __len__(self):
        return self.Element.__len__()

    def keys(self):
        #assert  self._children, 'Variable (dict) \'_children\' was not initialized  yet.'
        if self._children:
            return self._children.keys()
        return ''

    @property
    def has_children(self):
        return self.Element.__len__()

    @property
    def _get_children(self):
        return self.Element.getchildren()

    def _add_child(self, tagClass):
        self.Element.append(tagClass.Element)

    @property
    def has_param(self):
        try:
            return not self.Element.text.isspace()
        except:
            return False

    @property
    def param(self):
        if self.has_param:
            return self.Element.text
        return False

    @param.setter
    def param(self, txt_val):
        if txt_val is not str:
            txt_val = str(txt_val)
        self.Element.text = txt_val

    @property
    def has_attributes(self):
        return self.Element.attrib.__len__()

    def __str__(self):
        myStr = 'Object myXML_' + self.tag[0].capitalize() + self.tag[0:] + '\n' + \
        'tag: ' + self.tag + '\n' + \
        'param: ' + str(self.param) + '\n' + \
        'attributes: ' + self.attributes.__str__() + '\n'+ \
        'children: \n   '

        for key in self._children.keys():
            myStr = myStr + 'key: \'' + key + '\' ('
            child = self._children[key]
            if type(child) is list:
                if child.__len__() == 0:
                    myStr = myStr + '0'
                else:
                    myStr = myStr + str(child.__len__())
            else:
                myStr = myStr + '1'

            myStr = myStr + ')\n   '

        return myStr

    def __repr__(self):
        return self.__str__()


class myXML_AttributeHandler:
    """
    This is just a handler like a dict for. This is included in the myXML objects as a myXML.attributes
    """
    def __init__(self, AttrElement: ET.Element):
        #if AttrElement:
       self.Element = AttrElement

    def __len__(self):
        return self.Element.attrib.__len__()

    def __getitem__(self, item):
        #assert self.Element, 'There are no existing attributes.'
        if type(item) is int:
            assert item < self.__len__(), 'Index ' + str(item) + ' is higher than number of attributes. ' \
                                                                 'Number of attributes is ' + str(self.__len__())
            return self.Element.items(item)

        if type(item) is str:
            assert item in self.keys(), 'Attribute \'' + item + '\' does not exist for this instance.'
            return self.Element.attrib[item]

        return False

    def __setitem__(self, key, value):
        assert type(key) is str or isinstance(key, ET.QName), "Key for inserting an attribute must be string"
        #assert type(value) is str, "Attribute for xml parsing must be string"
        if not type(value): value = str(value)
        #if self.Element:
        self.Element.set(key, value)

    def __delitem__(self, key):
        assert type(key) is str, "Key for inserting an attribute must be string"
        del self.Element.attrib[key]

    def __str__(self):
        return self.Element.attrib.__str__()

    def __repr__(self):
        return self.__str__()

    def keys(self):
        return self.Element.keys()


# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import yaml

from datetime import datetime

from best.signal_generation.DCGAN.model import Discriminator, Generator, DiscriminatorTimeFreq
from best.signal_generation.DCGAN.dataset import DatasetTimeDomain
from best._config import ObjDictToDict

import os
import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt

from torch.utils.data import DataLoader
from best.signal import PSD


class Trainer:
    def __init__(self, config):
        self.cfg = config
        self.model_name = self.cfg.NAME
        exec(
        f"from {'.'.join(self.cfg.TRAIN.DATASET.split('.')[:-1])} import {self.cfg.TRAIN.DATASET.split('.')[-1]}\n" \
        f"self.DatasetTrain={self.cfg.TRAIN.DATASET.split('.')[-1]}({self.cfg.TRAIN.PATH_DATA}, categories_keep={self.cfg.TRAIN.CATEGORY_KEEP})"
        )
        exec(
        f"from {'.'.join(self.cfg.TRAIN.DATASET.split('.')[:-1])} import {self.cfg.TRAIN.DATASET.split('.')[-1]}\n" \
        f"self.DatasetTest={self.cfg.TRAIN.DATASET.split('.')[-1]}({self.cfg.TRAIN.PATH_DATA}, categories_keep={self.cfg.TRAIN.CATEGORY_KEEP})"
        )
        self.DatasetTrain.train()
        self.DatasetTest.eval()

        self.path_report = os.path.join(self.cfg.TRAIN.PATH_REPORT, self.cfg.NAME + "_" + datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))
        self.path_report_models = os.path.join(self.path_report, 'Models')
        self.path_report_images = os.path.join(self.path_report, 'Images')
        self.path_report_config = os.path.join(self.path_report, 'config.yaml')
        self.path_report_losses = os.path.join(self.path_report, 'losses.csv')
        self.path_report_losses_validation = os.path.join(self.path_report, 'losses_val.csv')

        os.mkdir(self.path_report)
        os.mkdir(self.path_report_models)
        os.mkdir(self.path_report_images)
        with open(self.path_report_config, 'w') as file:
            yaml.dump(ObjDictToDict(self.cfg), file)

        with open(self.path_report_losses, 'w') as f:
            f.write('Epoch, lr, loss_D_fake, loss_D_real, loss_G_discr, loss_G_spect\n')

        self.save_model_freq = self.cfg.TRAIN.SAVE_MODEL_EPOCH
        self.save_report_freq = self.cfg.TRAIN.SAVE_REPORT_EPOCH

        self.GPUs = self.cfg.TRAIN.GPU
        self.num_gpus = self.GPUs.__len__()
        self.minibatchsize = int(self.cfg.TRAIN.BATCH_SIZE)

        self.lr = self.cfg.TRAIN.BASE_LR

        self.epochs = self.cfg.TRAIN.EPOCHS


        self.current_epoch = 0
        self.current_iteration = 0
        self.overall_iteration = 0
        self.epoch_list = []
        self.lossKL_list = []
        self.lossMSE_list = []
        self.loss_list = []
        self.device = self.GPUs[0]

        self.Generator = Generator(n_features=self.cfg.MODEL.ARCHITECTURE.N_EMBEDDED_FEATURES, n_filters=self.cfg.MODEL.ARCHITECTURE.N_FILTERS)
        self.Discriminator = Discriminator(n_filters=self.cfg.MODEL.ARCHITECTURE.N_FILTERS)

        self.optimizerG = optim.Adam(self.Generator.parameters(), lr=self.cfg.TRAIN.LR, betas=(self.cfg.TRAIN.BETA_1, 0.999))
        self.optimizerD = optim.Adam(self.Discriminator.parameters(), lr=self.cfg.TRAIN.LR, betas=(self.cfg.TRAIN.BETA_1, 0.999))

        self.lr_schedulerG = torch.optim.lr_scheduler.MultiStepLR(self.optimizerG, milestones=self.cfg.TRAIN.DECAY_ITERATIONS, gamma=self.cfg.TRAIN.DECAY_FACTOR)
        self.lr_schedulerD = torch.optim.lr_scheduler.MultiStepLR(self.optimizerD, milestones=self.cfg.TRAIN.DECAY_ITERATIONS, gamma=self.cfg.TRAIN.DECAY_FACTOR)



        self.criterion = nn.BCELoss()
        self.criterion_MSE = nn.MSELoss()

        def weights_init(m):
            classname = m.__class__.__name__
            if classname.find('Conv') != -1:
                nn.init.normal_(m.weight.data, 0.0, 0.02)
            elif classname.find('BatchNorm') != -1:
                nn.init.normal_(m.weight.data, 1.0, 0.02)
                nn.init.constant_(m.bias.data, 0)

        self.Generator.apply(weights_init)
        self.Discriminator.apply(weights_init)

        self.DLoaderTrain = DataLoader(self.DatasetTrain, batch_size=self.minibatchsize, shuffle=True, num_workers=self.cfg.TRAIN.CPU_COUNT_LOADERS, drop_last=False)

        self.Discriminator.cuda(self.device)
        self.Generator.cuda(self.device)

    def do_epoch(self):
        self.current_iteration = 0
        print('----------------------------------')
        print('VAE Epoch ' + str(self.current_epoch))


        self.Discriminator.train()
        self.Generator.train()

        for k, x in enumerate(self.DLoaderTrain):
            x, y, ystr = x

            x = x.float().to(self.Discriminator.dummy_param.device)
            # break


            self.Discriminator.zero_grad()
            label = torch.full((x.shape[0] * 3,), 1.0, dtype=torch.float, device=self.Discriminator.dummy_param.device)
            # label = torch.full((x.shape[0],), 1.0, dtype=torch.float, device=self.Discriminator.dummy_param.device)
            output = self.Discriminator(x.unsqueeze(1)).view(-1)
            loss_D_real = self.criterion(output, label)
            loss_D_real.backward()


            noise = torch.randn(x.shape[0], 8, 3, device=self.Discriminator.dummy_param.device)
            fake = self.Generator(noise)
            label.fill_(0.0)
            output = self.Discriminator(fake.detach()).view(-1)
            loss_D_fake = self.criterion(output, label)
            loss_D_fake.backward()
            self.optimizerD.step()


            self.Generator.zero_grad()
            label.fill_(1.0)
            output = self.Discriminator(fake).view(-1)
            loss_G1 = self.criterion(output, label)


            x_orig_fft = torch.fft.fft(x.squeeze(1)).abs()
            x_fake_fft = torch.fft.fft(fake.squeeze(1)).abs()

            x_orig_fft = torch.log10(x_orig_fft+0.0001)
            x_fake_fft = torch.log10(x_fake_fft+0.0001)

            x_fake_fft[torch.isnan(x_orig_fft) | torch.isinf(x_orig_fft)] = 0
            x_orig_fft[torch.isnan(x_orig_fft) | torch.isinf(x_orig_fft)] = 0

            x_orig_fft[torch.isnan(x_fake_fft) | torch.isinf(x_fake_fft)] = 0
            x_fake_fft[torch.isnan(x_fake_fft) | torch.isinf(x_fake_fft)] = 0

            loss_G2 = self.criterion_MSE(x_fake_fft.mean(axis=0), x_orig_fft.mean(axis=0)) * 100

            loss_G = loss_G1 + loss_G2

            loss_G.backward()
            self.optimizerG.step()

            losses = {
                'loss_D_fake': loss_D_fake.item(),
                'loss_D_real': loss_D_real.item(),
                'loss_G_discr': loss_G1.item(),
                'loss_G_spect': loss_G2.item(),
            }


            if self.overall_iteration % self.cfg.TRAIN.SAVE_REPORT_ITERATION == 0 or k==0:
                self.print_losses_to_file(losses)
                self.plot_to_file(fake, x)

            if self.overall_iteration % self.cfg.TRAIN.SAVE_MODEL_ITERATION == 0 or k==0:
                self.save_model()

            self.current_iteration += 1
            self.overall_iteration += 1

        self.current_epoch += 1
        self.lr_schedulerD.step()
        self.lr_schedulerG.step()


    def print_losses_to_file(self, losses):
        path = self.path_report_losses
        lr = self.optimizerG.param_groups[0]['lr']
        #f.write('Epoch, lr, lossG_fake, lossG_real, lossD\n')
        printStr = f'{self.current_epoch}, {lr}, ' + ', '.join([f'{i:.8f}' for k, i in losses.items()]) + '\n'
        print(self.overall_iteration, self.current_epoch, self.current_iteration, losses)
        with open(path, 'a') as f:
            f.write(printStr)


    def plot_to_file(self, fake, real):
        x_real = real[-1].detach().cpu().squeeze().numpy()
        x_fake = fake[-1].detach().cpu().squeeze().numpy()
        img_path = os.path.join(self.path_report_images, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}")
        img_path_spect = os.path.join(self.path_report_images, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}_spect")

        plt.figure(figsize=(12, 4))
        ax0 = plt.subplot(2, 1, 1)
        plt.plot(x_real)
        plt.title('Real')
        plt.subplot(2, 1, 2, sharex=ax0)
        plt.plot(x_fake)
        plt.title('Fake')
        plt.savefig(img_path + '.png')
        plt.savefig(img_path + '.svg')
        plt.close()

        f, Pxx = PSD(fake.detach().cpu().numpy().squeeze(), 500)
        Pxx_fake = Pxx.mean(axis=0)

        f, Pxx = PSD(real.detach().cpu().numpy().squeeze(), 500)
        Pxx_real = Pxx.mean(axis=0)

        plt.figure(figsize=(12, 4))
        plt.semilogy(f[f<100], Pxx_real[f<100])
        plt.semilogy(f[f<100], Pxx_fake[f<100])
        plt.legend(['Real', 'Fake'])
        plt.savefig(img_path_spect + '.png')
        plt.savefig(img_path_spect + '.svg')
        plt.close()
        #plt.show()

    def save_model(self):
        PATH_D = os.path.join(self.path_report_models, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}_Discriminator")
        PATH_G = os.path.join(self.path_report_models, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}_Generator")
        torch.save(self.Discriminator.state_dict(), PATH_D)
        torch.save(self.Generator.state_dict(), PATH_G)

    def train(self):
        for k in range(self.epochs):
            self.do_epoch()


# class Trainer:
#     def __init__(self, config):
#         self.cfg = config
#         self.model_name = self.cfg.NAME
#         #self.DatasetTrain =
#         exec(
#         f"from {'.'.join(self.cfg.TRAIN.DATASET.split('.')[:-1])} import {self.cfg.TRAIN.DATASET.split('.')[-1]}\n" \
#         f"self.DatasetTrain={self.cfg.TRAIN.DATASET.split('.')[-1]}({self.cfg.TRAIN.PATH_DATA}, categories_keep={self.cfg.TRAIN.CATEGORY_KEEP})"
#         )
# 
#         self.path_report = os.path.join(self.cfg.TRAIN.PATH_REPORT, self.cfg.NAME + "_" + datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))
#         self.path_report_models = os.path.join(self.path_report, 'Models')
#         self.path_report_images = os.path.join(self.path_report, 'Images')
#         self.path_report_config = os.path.join(self.path_report, 'config.yaml')
#         self.path_report_losses = os.path.join(self.path_report, 'losses.csv')
#         self.path_report_losses_validation = os.path.join(self.path_report, 'losses_val.csv')
# 
#         os.mkdir(self.path_report)
#         os.mkdir(self.path_report_models)
#         os.mkdir(self.path_report_images)
#         with open(self.path_report_config, 'w') as file:
#             yaml.dump(ObjDictToDict(self.cfg), file)
# 
#         with open(self.path_report_losses, 'w') as f:
#             # f.write('Epoch, lr, loss_D_fake, loss_D_real, loss_DFFT_fake, loss_DFFT_real, loss_G_discr, loss_G_spect\n')
#             f.write('Epoch, lr, loss_D_fake_time, loss_D_real_freq, loss_D_fake_time, loss_D_freq, loss_G_time, loss_G_freq\n')
# 
#         self.save_model_freq = self.cfg.TRAIN.SAVE_MODEL_EPOCH
#         self.save_report_freq = self.cfg.TRAIN.SAVE_REPORT_EPOCH
# 
#         self.GPUs = self.cfg.TRAIN.GPU
#         self.num_gpus = self.GPUs.__len__()
#         self.minibatchsize = int(self.cfg.TRAIN.BATCH_SIZE)
# 
#         self.lr = self.cfg.TRAIN.BASE_LR
# 
#         self.epochs = self.cfg.TRAIN.EPOCHS
# 
# 
#         self.current_epoch = 0
#         self.current_iteration = 0
#         self.overall_iteration = 0
#         self.epoch_list = []
#         self.lossKL_list = []
#         self.lossMSE_list = []
#         self.loss_list = []
#         self.device = self.GPUs[0]
# 
#         self.Generator = Generator(n_features=self.cfg.MODEL.ARCHITECTURE.N_EMBEDDED_FEATURES, n_filters=self.cfg.MODEL.ARCHITECTURE.N_FILTERS)
#         self.Discriminator = DiscriminatorTimeFreq(n_filters=self.cfg.MODEL.ARCHITECTURE.N_FILTERS)
#         # self.DiscriminatorFFT = Discriminator(n_filters=self.cfg.MODEL.ARCHITECTURE.N_FILTERS)
# 
#         self.optimizerG = optim.Adam(self.Generator.parameters(), lr=self.cfg.TRAIN.LR, betas=(self.cfg.TRAIN.BETA_1, 0.999))
#         self.optimizerD = optim.Adam(self.Discriminator.parameters(), lr=self.cfg.TRAIN.LR, betas=(self.cfg.TRAIN.BETA_1, 0.999))
#         # self.optimizerDFFT = optim.Adam(self.DiscriminatorFFT.parameters(), lr=self.cfg.TRAIN.LR, betas=(self.cfg.TRAIN.BETA_1, 0.999))
# 
#         self.lr_schedulerG = torch.optim.lr_scheduler.MultiStepLR(self.optimizerG, milestones=self.cfg.TRAIN.DECAY_ITERATIONS, gamma=self.cfg.TRAIN.DECAY_FACTOR)
#         self.lr_schedulerD = torch.optim.lr_scheduler.MultiStepLR(self.optimizerD, milestones=self.cfg.TRAIN.DECAY_ITERATIONS, gamma=self.cfg.TRAIN.DECAY_FACTOR)
# 
# 
# 
#         self.criterion = nn.BCELoss()
#         self.criterion_MSE = nn.MSELoss()
# 
#         def weights_init(m):
#             classname = m.__class__.__name__
#             if classname.find('Conv') != -1:
#                 nn.init.normal_(m.weight.data, 0.0, 0.02)
#             elif classname.find('BatchNorm') != -1:
#                 nn.init.normal_(m.weight.data, 1.0, 0.02)
#                 nn.init.constant_(m.bias.data, 0)
# 
#         self.Generator.apply(weights_init)
#         self.Discriminator.apply(weights_init)
#         # self.DiscriminatorFFT.apply(weights_init)
# 
#         self.DLoaderTrain = DataLoader(self.DatasetTrain, batch_size=self.minibatchsize, shuffle=True, num_workers=self.cfg.TRAIN.CPU_COUNT_LOADERS, drop_last=False)
# 
#         self.Discriminator.cuda(self.device)
#         # self.DiscriminatorFFT.cuda(self.device)
#         self.Generator.cuda(self.device)
# 
#     def do_epoch(self):
#         self.current_iteration = 0
#         print('----------------------------------')
#         print('VAE Epoch ' + str(self.current_epoch))
# 
# 
#         self.Discriminator.train()
#         # self.DiscriminatorFFT.train()
#         self.Generator.train()
# 
#         for k, x in enumerate(self.DLoaderTrain):
#             x, y, ystr = x
#             # break
# 
#             x = x.float().to(self.Discriminator.dummy_param.device)
#             x_orig_fft = torch.fft.fft(x.squeeze(1)).abs()#[:, :750]
#             x_orig_fft = torch.log10(x_orig_fft + 1e-12)
# 
#             # x = x.unsqueeze(1)
# 
#             self.Discriminator.zero_grad()
#             label = torch.full((x.shape[0] * 1,), 1.0, dtype=torch.float, device=self.Discriminator.dummy_param.device)
#             # label = torch.full((x.shape[0],), 1.0, dtype=torch.float, device=self.Discriminator.dummy_param.device)
#             output_time, output_freq = self.Discriminator((x.unsqueeze(1), x_orig_fft.unsqueeze(1)))#.view(-1)
#             loss_D_real_time = self.criterion(output_time.view(-1), label) * 0.8
#             loss_D_real_freq = self.criterion(output_freq.view(-1), label) * 0.2
#             loss_D_real = loss_D_real_time + loss_D_real_freq
#             loss_D_real.backward()
# 
#             # self.DiscriminatorFFT.zero_grad()
#             # label = torch.full((x_orig_fft.shape[0] * 3,), 1.0, dtype=torch.float, device=self.Discriminator.dummy_param.device)
#             # # label = torch.full((x.shape[0],), 1.0, dtype=torch.float, device=self.Discriminator.dummy_param.device)
#             # output = self.DiscriminatorFFT(x.unsqueeze(1)).view(-1)
#             # loss_DFFT_real = self.criterion(output, label)
#             # loss_DFFT_real.backward()
# 
#             noise = torch.randn(x.shape[0], self.cfg.MODEL.ARCHITECTURE.N_EMBEDDED_FEATURES, 3, device=self.Discriminator.dummy_param.device)
#             noise[:, :, 1] = noise[:, :, 0] + (0.5 * noise[:, :, 1])
#             noise[:, :, 2] = noise[:, :, 1] + (0.5 * noise[:, :, 2])
# 
#             fake = self.Generator(noise)#.detach()
#             x_fake_fft = torch.fft.fft(fake.squeeze(1)).abs()#[:, :750]
#             x_fake_fft = torch.log10(x_fake_fft + 1e-12)
# 
#             label.fill_(0.0)
#             output_time, output_freq = self.Discriminator((fake.detach(), x_fake_fft.detach().unsqueeze(1)))
#             loss_D_fake_time = self.criterion(output_time.view(-1), label) * 0.8
#             loss_D_fake_freq = self.criterion(output_freq.view(-1), label) * 0.2
#             loss_D_fake = loss_D_fake_time + loss_D_fake_freq
#             loss_D_fake.backward()
#             # output = self.DiscriminatorFFT(x_fake_fft.detach().unsqueeze(1)).view(-1)
#             # loss_DFFT_fake = self.criterion(output, label)
#             # loss_DFFT_fake.backward()
# 
#             self.optimizerD.step()
#             # self.optimizerDFFT.step()
# 
# 
#             self.Generator.zero_grad()
#             label.fill_(1.0)
#             output_time, output_freq = self.Discriminator((fake, x_fake_fft.unsqueeze(1)))
#             # outputFFT = self.DiscriminatorFFT(x_fake_fft.unsqueeze(1)).view(-1)
# 
#             loss_G1 = self.criterion(output_time.view(-1), label) * 0.8
#             loss_G2 = self.criterion(output_freq.view(-1), label) * 0.2
# 
# 
#             # x_orig_fft[torch.isnan(x_orig_fft) | torch.isinf(x_orig_fft)] = 0
#             # x_orig_fft[torch.isnan(x_fake_fft) | torch.isinf(x_fake_fft)] = 0
# 
# 
#             loss_G = loss_G1 + loss_G2
#             loss_G.backward()
#             self.optimizerG.step()
# 
#             losses = {
#                 'loss_D_fake_time': loss_D_fake_time.item(),
#                 'loss_D_real_time': loss_D_real_time.item(),
#                 'loss_D_fake_freq': loss_D_fake_freq.item(),
#                 'loss_D_real_freq': loss_D_real_freq.item(),
#                 # 'loss_DFFT_fake': loss_DFFT_fake.item(),
#                 # 'loss_DFFT_real': loss_DFFT_real.item(),
#                 'loss_G_time': loss_G1.item(),
#                 'loss_G_spect': loss_G2.item(),
#             }
# 
# 
#             if self.overall_iteration % self.cfg.TRAIN.SAVE_REPORT_ITERATION == 0 or k==0:
#                 self.print_losses_to_file(losses)
#                 self.plot_to_file(fake, x)
# 
#             if self.overall_iteration % self.cfg.TRAIN.SAVE_MODEL_ITERATION == 0 or k==0:
#                 self.save_model()
# 
#             self.current_iteration += 1
#             self.overall_iteration += 1
# 
#         self.current_epoch += 1
#         self.lr_schedulerD.step()
#         self.lr_schedulerG.step()
# 
# 
#     def print_losses_to_file(self, losses):
#         path = self.path_report_losses
#         lr = self.optimizerG.param_groups[0]['lr']
#         #f.write('Epoch, lr, lossG_fake, lossG_real, lossD\n')
#         printStr = f'{self.current_epoch}, {lr}, ' + ', '.join([f'{i:.8f}' for k, i in losses.items()]) + '\n'
#         print(self.overall_iteration, self.current_epoch, self.current_iteration, losses)
#         with open(path, 'a') as f:
#             f.write(printStr)
# 
# 
#     def plot_to_file(self, fake, real):
#         x_real = real[-1].detach().cpu().squeeze().numpy()
#         x_fake = fake[-1].detach().cpu().squeeze().numpy()
#         img_path = os.path.join(self.path_report_images, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}")
#         img_path_spect = os.path.join(self.path_report_images, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}_spect")
# 
#         plt.figure(figsize=(12, 4))
#         ax0 = plt.subplot(2, 1, 1)
#         plt.plot(x_real)
#         plt.title('Real')
#         plt.subplot(2, 1, 2, sharex=ax0)
#         plt.plot(x_fake)
#         plt.title('Fake')
#         plt.savefig(img_path + '.png')
#         plt.savefig(img_path + '.svg')
#         plt.close()
# 
#         f, Pxx = PSD(fake.detach().cpu().numpy().squeeze(), 500)
#         Pxx_fake = Pxx.mean(axis=0)
# 
#         f, Pxx = PSD(real.detach().cpu().numpy().squeeze(), 500)
#         Pxx_real = Pxx.mean(axis=0)
# 
#         plt.figure(figsize=(12, 4))
#         plt.semilogy(f[f<100], Pxx_real[f<100])
#         plt.semilogy(f[f<100], Pxx_fake[f<100])
#         plt.legend(['Real', 'Fake'])
#         plt.savefig(img_path_spect + '.png')
#         plt.savefig(img_path_spect + '.svg')
#         plt.close()
#         #plt.show()
# 
#     def save_model(self):
#         PATH_D = os.path.join(self.path_report_models, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}_Discriminator")
#         PATH_G = os.path.join(self.path_report_models, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}_Generator")
#         torch.save(self.Discriminator.state_dict(), PATH_D)
#         torch.save(self.Generator.state_dict(), PATH_G)
# 
#     def train(self):
#         for k in range(self.epochs):
#             self.do_epoch()



# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import os
import numpy as np

from copy import deepcopy
from shutil import rmtree

import unittest
from unittest import TestCase

import torch.cuda

from best.deep_learning.seizure_detect import load_trained_model, preprocess_input, infer_seizure_probability
from numpy.random import rand
from best.deep_learning.models_generic import *


class Test_deep_learning(TestCase):
    def test_import(self):
        print("Testing import 'from best.deep_learning'" )


class Test_seizure_detect(TestCase):
    def test_load_trained_model(self):
        modelA = load_trained_model('modelA')
        modelB = load_trained_model('modelB')

    def test_preprocess_input(self):
        modelA = load_trained_model('modelA')
        fs = 500
        x_len = 300
        channels = 1
        x_input = rand(channels, fs * x_len)
        x = preprocess_input(x_input, fs)
        y = infer_seizure_probability(x, modelA)
        self.assertEqual(599, y.shape[1])
        channels = 3
        x_input = rand(channels, fs * x_len)
        x = preprocess_input(x_input, fs)
        y = infer_seizure_probability(x, modelA)
        self.assertEqual(599, y.shape[1])
        if torch.cuda.is_available():
            modelA.cuda(0)
            y = infer_seizure_probability(x, modelA, True, 0)
            self.assertEquals(599, y.shape[1])



if __name__ == '__main__':
    unittest.main()



from best.dbs.artifact_removal._configs import configs_ArtifactEraser
from best.dbs.artifact_removal._models import models_ArtifactEraser
from best.dbs.artifact_removal._remove_stimulation_artifacts import remove_artifacts



# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


from best import DELIMITER
from best._config import get_files
from best.signal_generation.DCGAN._configs import configs_DCGAN
from best.signal_generation.DCGAN.model import Generator
import torch


class _models_DCGAN(dict):

    _keys = dict([
        (
            '_'.join(f.split(DELIMITER)[-1].split('_')[:2]),
            f
        )
        for f in get_files(DELIMITER.join(__file__.split(DELIMITER)[:-1]), 'pt')
        if not '.-' in f.split(DELIMITER)[-1]
    ])

    def keys(self):
        return self._keys.keys()

    def get_model(self, item):
        f = self._keys[item]
        cfg = configs_DCGAN[item]
        state_dict = torch.load(f, map_location='cpu')
        mod = Generator(n_features = cfg.MODEL.ARCHITECTURE.N_EMBEDDED_FEATURES, n_filters = cfg.MODEL.ARCHITECTURE.N_FILTERS)
        mod.load_state_dict(state_dict, strict=True)
        mod.eval()
        return mod

    def __getitem__(self, item):
        return self.get_model(item)

    def __call__(self, item):
        return self[item]

    def __str__(self):
        return 'models_DCGAN: ' + str(self._keys.keys())

    def __repr__(self):
        return self.__str__()

models_DCGAN = _models_DCGAN()








# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import os
import numpy as np

from copy import deepcopy
from shutil import rmtree

import unittest
from unittest import TestCase


# define env variables for the db testing
import os
os.environ['IP_SQL'] = '0.0.0.0'
os.environ['USER_SQL'] = 'db_usr'
os.environ['PW_SQL'] = '\'db_pwd\''
os.environ['PORT_SQL'] = '3360'
os.environ['IP_SSH'] = '\'0.0.0.0\''
os.environ['USER_SSH'] = '\'ssh_user\''
os.environ['PW_SSH'] = '\'ssh_pwd\''
os.environ['PORT_SSH'] = '\'22\''
os.environ['IP_MEF'] = '\'0.0.0.0\''
os.environ['PORT_MEF'] = '\'[5500,5501,5502]\''

# from best.cloud.db import *



class TestFeature(TestCase):
    def test_import_mefclient_connection_variables(self):
        print("Testing import from best.cloud._mefclient_connection_variables import *'")
        from best.cloud._mefclient_connection_variables import IP_MEF, PORT_MEF
        self.assertTrue(IP_MEF == '\'0.0.0.0\'')
        self.assertTrue(PORT_MEF == '\'[5500,5501,5502]\'')

    def test_import_mefclient_connection_variables(self):
        print("Testing import 'from best.cloud._db_connection_variables import *'")
        from best.cloud._db_connection_variables import IP_SQL, PORT_SQL, PW_SQL, USER_SQL, IP_SSH, PORT_SSH, PW_SSH, USER_SSH
        self.assertTrue(IP_SQL == '0.0.0.0')
        self.assertTrue(PORT_SQL == '3360')
        self.assertTrue(PW_SQL == 'db_pwd')
        self.assertTrue(USER_SQL == 'db_usr')
        self.assertTrue(USER_SSH == 'ssh_user')
        self.assertTrue(IP_SSH == '0.0.0.0')
        self.assertTrue(PW_SSH == 'ssh_pwd')
        self.assertTrue(PORT_SSH == '22')

if __name__ == '__main__':
    unittest.main()



# Copyright 2020-present, Mayo Clinic Department of Neurology - Bioelectronics Neurophysiology and Engineering Laboratory
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import os

def validate_environment_db_variables():
    for key in sql_vars:
        if not key in environ:
            raise KeyError('[ENVIRONMENT VARIABLE MISSING]: Environment variable ' + key + ' was not found. '
                                                                                           'Following env variables required for a succesfull database connection setup: ' + str(sql_vars))


    w_raised = False
    for key in sql_vars:
        if not key in environ:
            raise Warning('[ENVIRONMENT VARIABLE MISSING]: Environment variable ' + key + ' was not found.')
            w_raised = True
    if w_raised:
        raise Warning('[SSH CONNECTION IGNORED]: Connection details for SSH connection were not found. DatabaseHandler will'
                      'connect directly to database from your current IP address.')

# SQL variables
IP_SQL = USER_SQL = PW_SQL = PORT_SQL = None
sql_vars = ['IP_SQL', 'USER_SQL', 'PW_SQL', 'PORT_SQL']
# SSH_variables
IP_SSH = USER_SSH = PW_SSH = PORT_SSH = None
ssh_vars = ['IP_SSH', 'USER_SSH', 'PW_SSH', 'PORT_SSH']


environ = list(dict(os.environ).keys())
validate_environment_db_variables()


for sqlv in sql_vars:
    var = os.environ[sqlv]
    if not '\'' in var: var = '\'' + var + '\''
    exec(sqlv + '=' + var.replace(' ', ''))

for sshv in ssh_vars:
    if sshv in environ:
        var = os.environ[sshv]
        if not '\'' in var: var = '\'' + var + '\''
        exec(sshv + '=' + var.replace(' ', ''))
    else:
        exec(sshv + '=' + 'None')



# Copyright 2020-present, Mayo Clinic Department of Neurology - Laboratory of Bioelectronics Neurophysiology and Engineering
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


"""
Python Invasive Electrophysiology Signal Processing Toolbox - sleep_classification (c) is a Python package developed by the `Bioelectronics Neurophysiology and Engineering <https://www.mayo.edu/research/labs/bioelectronics-neurophysiology-engineering/overview>`_ laboratory of Department of Neurology Mayo Clinic.
The main scope of this package is to provide tools for an automated sleep classification using intracranial EEG signals.
"""

__version__ = '0.1.5'



import os
# Check windows or linux and sets separator
if os.name == 'nt': DELIMITER = '\\'
else: DELIMITER = '/'














from best import DELIMITER
from best._config import get_files, config

configs_ArtifactEraser = dict([(f.split(DELIMITER)[-1].split('.')[0], config(f)) for f in get_files(DELIMITER.join(__file__.split(DELIMITER)[:-1]), 'yaml') if not '.-' in f.split(DELIMITER)[-1]])








# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
import numpy as np

from sklearn.decomposition import PCA
from best.feature import augment_features

"""
Modules compatioble with scikitlearn pipeline

"""

class FeatureAugmentorModule:
    """
    Feature augmentation using an 'augment_features' function from the 'PiesUtils' package. See the code for additional details.
    """
    def __init__(self):
        self.mutual_functions = [np.divide, np.multiply]
        self.mutual_function_symbol = ['*', '/']

        self.standalone_functions = []
        self.mutual_function_symbol = []

    def fit(self, X=None, Y=None):
        pass

    def transform(self, X):
        n_orig_features = X.shape[1]
        for idx, func in enumerate(self.mutual_functions):
            X = augment_features(X, operation=func, mutual=True, feature_indexes=np.arange(n_orig_features))

        for idx, func in enumerate(self.standalone_functions):
            X = augment_features(X, operation=func, mutual=False, feature_indexes=np.arange(n_orig_features))
        return X

    def fit_transform(self, X, Y=None):
        return self.transform(X)

    def __call__(self, X):
        return self.transform(X)


class ZScoreModule:
    """
    Z-score normalization compatible with scikit.pipeline.Pipeline
    Enables continuous learning - enabling continuous adaptation.

    Modes
        * Zscore normalization
        * Zscore normalization with fixed mean and std values based on the initial training dataset
            * Possible category-wise normalization with mean and std values estimated from the training dataset - number of features is multiplied by number of categories
        * Zscore normalization with an initial mean and std values trained on the training dataset - adaptation during inference
            https://stats.stackexchange.com/questions/211837/variance-of-subsample



    Attributes
    ----------
    continuous_learning : bool
            If true - An instance updates mean and variance values during each prediction step. Initial outlier filtering is recommended
    trainable : bool
        If false - An instance normalizes inference data based on their current mean value and std
        If true - An instance remembers mean and variance values of training data
    multi_class : bool
        If true - An instance performs normalization for each training class separately
        Number of output features is multiplied by a number of training categories
    mean : numpy ndarray / list
        Trained mean values for each feature. In case multi_class == True -> list of numpy ndarrays for each category
    std : numpy ndarray

    N : int

    """
    def __init__(self, trainable=False, continuous_learning=False, multi_class=False):
        """
        Parameters
        ----------
        continuous_learning : bool
            If true - An instance updates mean and variance values during each prediction step. Initial outlier filtering is recommended
        trainable : bool
            If false - An instance normalizes inference data based on their current mean value and std
            If true - An instance remembers mean and variance values of training data
        multi_class : bool
            If true - An instance performs normalization for each training class separately
            Number of output features is multiplied by a number of training categories
        """


        self.continuous_learning = continuous_learning
        self.trainable = trainable
        self.multi_class = multi_class

        #self.sum_square = None
        self.mean_square = None
        self.mean = None
        self.std = None
        self.N = 0
        self.categories = []

        self._check_modes()

    def fit(self, X=None, Y=None):
        """
        Parameters
        ----------
        X : numpy ndarray
            shape[n_samples, n_features]
        Y : list or numpy array, optional
            category reference for each sample - required only for option with multi_class normalization

        Returns
        --------
        None

        """
        self._check_modes()
        if self.trainable is True:
            if self.multi_class is True:
                self.mean = []
                self.std = []
                self.categories = np.unique(Y)
                Y = np.array(Y)
                for cat_key in self.categories:
                    X_ = X[Y == cat_key]
                    self.mean.append(X_.mean(axis=0).reshape(1, -1))
                    self.std.append(X_.std(axis=0).reshape(1, -1))

            else:
                self.mean = X.mean(axis=0).reshape(1, -1)
                self.std = X.std(axis=0).reshape(1, -1)
                self.N = X.shape[0]
                #self.sum_square = (X**2).sum(axis=0).reshape(1, -1)
                self.mean_square = (X**2).mean(axis=0).reshape(1, -1)


    def transform(self, X=None):
        """
        Parameters
        ----------
        X : numpy ndarray
            shape[n_samples, n_features]

        Returns
        -------
        transformed_data : numpy ndarray
            shape[n_samples, n_features]

        """
        self._check_modes()
        if self.trainable is True:
            if self.multi_class is True:
                temp_list = []
                for cat_idx in range(self.mean.__len__()):
                    mean = self.mean[cat_idx]
                    std = self.std[cat_idx]
                    X_temp = (X - mean) / std
                    temp_list.append(X_temp)
                return np.concatenate(temp_list, axis=1)

            elif self.continuous_learning is True:
                M = X.shape[0]
                N = self.N

                mean_new = X.mean(axis=0).reshape(1, -1)
                mean_old = self.mean

                std_new = X.std(axis=0).reshape(1, -1)
                std_old = self.std

                #sum_square_new = (X**2).sum(axis=0).reshape(1, -1)
                #sum_square_old = self.sum_square

                mean_square_new = (X**2).mean(axis=0).reshape(1, -1)
                mean_square_old = self.mean_square


                c_old = N / (M+N)
                c_new = M / (M+N)
                mean_joint = (mean_old * c_old) + (mean_new * c_new)
                #mean_square_joint = mean_square_old * c_old + mean_square_new * c_new
                mean_square_joint = (mean_square_old * N + mean_square_new * M) / (M + N)

                #sum_square_joint = sum_square_new + sum_square_old

                #std_joint = ((std_old * N) + (std_new * M)) / (M+N) # just average std
                #std_joint = np.sqrt((N * (std_old**2 + (mean_old - mean_joint)**2)) +  (M * (std_new**2 + (mean_new - mean_joint)**2))) / (M + N)
                std_joint = np.sqrt(
                    (
                            N*std_old**2 + M*std_new**2 + N*((mean_old - mean_joint)**2) + M*((mean_new - mean_joint)**2)
                    ) / (N+M)
                )


                #std_joint = np.sqrt(((sum_square_joint) / (N + M)) - mean_joint**2) # average std using sum_square

                #std_joint = np.sqrt((mean_square_joint) - mean_joint**2) # average std using sum_square
                #std_joint = np.sqrt(mean_square_joint - mean_joint**2)

                self.mean = mean_joint
                self.std = std_joint
                self.N = N + M

                return (X - self.mean) / self.std

            else: # trainable
                return (X - self.mean) / self.std

        return (X - X.mean(axis=0).reshape(1, -1)) / X.std(axis=0).reshape(1, -1) # just zscore

    def fit_transform(self, X=None, Y=None):
        """
        Parameters
        ----------
        X : numpy ndarray
            shape[n_samples, n_features]
        Y : list or numpy array, optional
            category reference for each sample - required only for option with multi_class normalization

        Returns
        -------
        transformed_data : numpy ndarray
            shape[n_samples, n_features]

        """
        self.fit(X, Y)
        return self.transform(X)


    def __call__(self, X=None, Y=None):
        return self.transform(X)

    def _check_modes(self):
        if self.trainable is True:
            if self.continuous_learning is True and self.multi_class is True:
                raise AssertionError('[ASSERTION ERROR] - ZScore module - only one of parameters continuous_learning, multi_class can be activated')
        if self.trainable is False:
            if self.continuous_learning is True or self.multi_class is True:
                raise Warning('[WARNING] - ZScore - if trainable is False all other parameters will be considered as False')


class LogModule:
    def __init__(self):
        pass

    def fit(self, X, Y=None):
        pass

    def transform(self, X, Y=None):
        return np.log(X)

    def fit_transform(self, X, Y=None):
        return self.transform(X)

    def __call__(self, X):
        return self.transform(X)


class Log10Module:
    def __init__(self):
        pass

    def fit(self, X, Y=None):
        pass

    def transform(self, X, Y=None):
        return np.log10(X)

    def fit_transform(self, X, Y=None):
        return self.transform(X)

    def __call__(self, X):
        return self.transform(X)


class PCAModuleSVD:

    def __init__(self, var_threshold = 0.98):
        self.var_threshold = var_threshold

    def fit(self, X, Y=None):
        # Data matrix X, assumes 0-centered
        n, m = X.shape
        #assert np.allclose(X.mean(axis=0), np.zeros(m))
        # Compute covariance matrix
        C = np.dot(X.T, X) / (n-1)
        # Eigen decomposition
        self.eigen_vals, self.eigen_vecs = np.linalg.eig(C)

        norm_eigs = self.eigen_vals / self.eigen_vals.sum()
        self.n = 1
        while norm_eigs[:self.n].sum() < self.var_threshold:
            self.n += 1

    def transform(self, X, Y=None):
        return np.dot(X, self.eigen_vecs[:, :self.n])


    def fit_transform(self, X, Y=None):
        self.fit(X)
        return self.transform(X)

    def __call__(self, X):
        return self.transform(X)


class PCAModule(PCA):
    def __init__(self, var_threshold = 0.98):
        super().__init__()
        self.var_threshold = var_threshold

    def fit(self, X, y=None):
        testPCA = PCA()
        testPCA.fit(X)
        n = 1
        vals = testPCA.singular_values_
        vals = vals / vals.sum()
        while vals[:n].sum() < self.var_threshold:
            n += 1

        super().__init__(n_components=n)
        super().fit(X, y)
        return self

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X)




from best import DELIMITER
from best._config import get_files
from best.dbs.artifact_removal._configs import configs_ArtifactEraser
from best.dbs.artifact_removal.model import *
import torch


class _models_ArtifactEraser(dict):

    _keys = dict([
        (
            # '_'.join(f[:-3].split(DELIMITER)[-1].split('_')[:3]),
            f[:-3].split(DELIMITER)[-1],
            f
        )
        for f in get_files(DELIMITER.join(__file__.split(DELIMITER)[:-1]), 'pt')
        if not '.-' in f.split(DELIMITER)[-1]
    ])

    def keys(self):
        return self._keys.keys()

    def get_model(self, item):
        f = self._keys[item]
        cfg = configs_ArtifactEraser[item]
        state_dict = torch.load(f, map_location='cpu')
        #mod = None
        # mod = GRU_Denoiser(n_filters = cfg.MODEL.ARCHITECTURE.N_FILTERS)
        exec(f"self.mod={cfg.MODEL.MODEL.split('.')[-1]}(n_filters={cfg.MODEL.ARCHITECTURE.N_FILTERS})")
        self.mod.load_state_dict(state_dict, strict=True)
        self.mod.eval()
        mod = self.mod
        del self.mod
        return mod

    def __getitem__(self, item):
        return self.get_model(item)

    def __call__(self, item):
        return self[item]

    def __str__(self):
        return 'models_ArtifactEraser: ' + str(self._keys.keys())

    def __repr__(self):
        return self.__str__()

models_ArtifactEraser = _models_ArtifactEraser()









# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import yaml
import os
import numpy as np


class ObjDict(dict):
    """
    Dictionary which you can access a) as a dict; b) as a struct with attributes. Can use both foo adding and deleting
    attributes resp items. Inherits from dict
    """

    def __init__(self, VT_={}):
        super().__init__(VT_)
        for key in VT_.keys():
            self.__setattr__(key, VT_[key])

    def __setitem__(self, key, value):
        if key in self:
            del self[key]

        super().__setitem__(key, value)
        if not key in self.__dir__():
            self.__setattr__(key, value)

    def __setattr__(self, key, value):
        if key in self.__dir__():
            self.__delattr__(key)

        super().__setattr__(key, value)
        if not key in self:
            self.__setitem__(key, value)

    def __delitem__(self, key):
        value = super().pop(key)
        try:
            super().pop(value, None)
        except: pass
        if key in dir(self):
            self.__delattr__(key)

    def __delattr__(self, key):
        super().__delattr__(key)
        if key in self:
            self.__delitem__(key)

    def __repr__(self):
        return f"{type(self).__name__}({super().__repr__()})"

    def __missing__(self, key):
        self[key] = ObjDict()
        return self[key]

    def __getattr__(self, item):
        if not item in self.__dir__():
            self.__missing__(item)
        return super().__getattribute__(item)


def DictToObjDict(d):
    """
    Converts dictionary into object where, keys - converts keys into attributes
    """
    if isinstance(d, dict):
        #print(d.keys())
        top = ObjDict(d)

        #print(top.keys())
        for key in d.keys():
            if isinstance(d[key], dict):
                tmp = DictToObjDict(d[key])
                del top[key]
                top[key] = tmp

        return top
    else: return d


def config(path_config):
    with open(path_config, 'r') as stream:
        Cfg = yaml.safe_load(stream)
    return DictToObjDict(Cfg)


def get_files(path, endings=None, creation_time=False):
    """File list generator.

        For each file in path directory and subdirectories with endings from endings_tuple returns path in a list variable

        path is a string, the path to the directory where you wanna search specific files.

        endings_tuple is a tuple of any length

        The function returns paths to all files in folder and files in all subfolders as well.

        Caution:  ending must be tuple, even for length of 1 ending. ('png')

        Example:

        from myFiles import get_files
        import pandas as pd

        path = "root/data"
        files_list = get_files(path)
        files_pd = pd.DataFrame(files_list)

        def get_FID(x):
            temp = x['path'].split('\\')
            return temp[len(temp) - 1][0:-4]
        files_pd = files_pd.rename(columns={0: "path"})
        files_pd['fid'] = files_pd.apply(lambda x: get_FID(x), axis=1)
        """

    if isinstance(endings, str):
        endings = (endings, )

    if isinstance(endings, type(None)):
        data = [os.path.join(root, name)
                for root, dirs, files in os.walk(path)
                for name in files]

        data += [os.path.join(root, dir)
                for root, dirs, files in os.walk(path)
                for dir in dirs]
    else:
        data = [os.path.join(root, name)
                for root, dirs, files in os.walk(path)
                for name in files
                if name.endswith(endings)]

        data += [os.path.join(root, dir)
                for root, dirs, files in os.walk(path)
                for dir in dirs
                if dir.endswith(endings)]


    data = [file for file in data if not '._' in file]
    data.sort()

    if creation_time == False:
        return data

    creation_times = [os.stat(path).st_ctime for path in data]
    data = np.array(data)
    creation_times = np.array(creation_times)
    idx = creation_times.argsort()
    sorted_data = data[idx]
    sorted_creation_times = creation_times[idx]
    return sorted_data, sorted_creation_times


def ObjDictToDict(d):
    """
    Converts dictionary into object where, keys - converts keys into attributes
    """
    if isinstance(d, ObjDict):
        #print(d.keys())
        top = dict(d)

        #print(top.keys())
        for key in d.keys():
            if isinstance(d[key], ObjDict):
                tmp = ObjDictToDict(d[key])
                del top[key]
                top[key] = tmp

        return top
    else: return d





# Copyright 2020-present, Mayo Clinic Department of Neurology - Bioelectronics Neurophysiology and Engineering Laboratory
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import os

def validate_environment_variables():
    for key in mef_vars:
        if not key in environ:
            raise KeyError('[ENVIRONMENT VARIABLE MISSING]: Environment variable ' + key + ' was not found. '
                                                                                           'Following env variables required for a succesfull database connection setup: ' + str(mef_vars))



# SQL variables
PORT_MEF = IP_MEF = None
mef_vars = ['PORT_MEF', 'IP_MEF']



environ = list(dict(os.environ).keys())
validate_environment_variables()


for mefv in mef_vars:
    var = os.environ[mefv]

    if not '\'' in var: var = '\'' + var + '\''

    if mefv == 'PORT_MEF': # erase ' to create list and not string
        var = var.replace('\'', '')
    exec(mefv + '=' + var.replace(' ', ''))



# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.



# Copyright 2020-present, Mayo Clinic Department of Neurology - Bioelectronics Neurophysiology and Engineering Laboratory
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import zmq
import numpy as np
from best.cloud import *
from best.cloud._mefclient_connection_variables import *

class MefClient:
    __version__ = '3.0.0'
    def __init__(self,ports=None, server_ip=None, response_wait=30):
        if isinstance(ports, type(None)):
            ports = PORT_MEF
        if isinstance(server_ip, type(None)):
            server_ip = IP_MEF

        context = zmq.Context(1)
        self.client = context.socket(zmq.REQ)
        self.poll = zmq.Poller()
        self.ports = ports
        self.server_ip = server_ip
        self.response_wait = response_wait

    def request_data(self, path, channel, passwd='', start=None, stop=None, sample=0):
        a, b, c = self._request_data(path, channel, passwd, start, stop, sample)
        if a: return a, b, c
        a, b, c = self._request_data(path, channel, passwd, start, stop, sample)
        if a: return a, b, c
        a, b, c = self._request_data(path, channel, passwd, start, stop, sample)
        return a, b, c


    def _request_data(self, path, channel, passwd='', start=None, stop=None, sample=0):
        #if start and stop:
            #if np.log10(start) < 15:
                #start = int(round(start * 1e6))
                #stop = int(round(stop * 1e6))

        # sample arg must be 1 if data are read by sample
        client = self.client
        for p in self.ports:
            client.connect(f"tcp://{self.server_ip}:{p}")

        self.poll.register(client, zmq.POLLIN)
        try:
            client.send_json({'path': path, 'channel': channel, 'passwd': passwd, 'start': start, 'stop': stop, 'sample': sample,
                              'flag':'data'})
            socks = dict(self.poll.poll(self.response_wait * 1000))
            if socks.get(client) == zmq.POLLIN:
                md = client.recv_json(flags=0, )
                msg = client.recv(flags=0, copy=True, track=False)
                buf = memoryview(msg)
                if md['error'] == 1:
                    return False, md['error_message'], False
                else:
                    res_numpy = np.frombuffer(buf, dtype=md['dtype'])
                    return True, res_numpy.reshape(md['shape']), md['fsamp']
            else:
                client.setsockopt(zmq.LINGER, 0)
                client.close()
                self.poll.unregister(client)
                # rebind
                context = zmq.Context(1)
                self.client = context.socket(zmq.REQ)
                return False, f'Server response time elapsed {self.response_wait} s', False

        except Exception as exc:
            exce = f'{exc}'
            return False, exce, False

    def request_processed_data(self, path, channel, passwd='', start=None, stop=None, sample=0, b_low=None, a_low=None, b_high=None, a_high=None, zero_phase=True, fs_target=200, correct_timestamp=True):
        #if start and stop:
            #if np.log10(start) < 15 and correct_timestamp:
                #start = int(round(start * 1e6))
                #stop = int(round(stop * 1e6))

        # sample arg must be 1 if data are read by sample
        client = self.client
        for p in self.ports:
            client.connect(f"tcp://{self.server_ip}:{p}")

        self.poll.register(client, zmq.POLLIN)
        try:
            client.send_json({
                'path': path,
                'channel': channel,
                'passwd': passwd,
                'start': int(start),
                'stop': int(stop),
                'sample': int(sample),
                'flag':'data',
                'filter_b_high': [float(v) for v in b_high],
                'filter_a_high': [float(v) for v in a_high],
                'filter_b_low': [float(v) for v in b_low],
                'filter_a_low': [float(v) for v in a_low],
                'zero_phase': bool(zero_phase),
                'fs_target': float(fs_target)
            })


            socks = dict(self.poll.poll(self.response_wait * 1000))
            if socks.get(client) == zmq.POLLIN:
                md = client.recv_json(flags=0, )
                msg = client.recv(flags=0, copy=True, track=False)
                buf = memoryview(msg)
                if md['error'] == 1:
                    return False, md['error_message'], False
                else:
                    res_numpy = np.frombuffer(buf, dtype=md['dtype'])
                    return True, res_numpy.reshape(md['shape']), md['fsamp']
            else:
                client.setsockopt(zmq.LINGER, 0)
                client.close()
                self.poll.unregister(client)
                # rebind
                context = zmq.Context(1)
                self.client = context.socket(zmq.REQ)
                return False, f'Server response time elapsed {self.response_wait} s', False

        except Exception as exc:
            exce = f'{exc}'
            return False, exce, False




    def request_data_bulk(self, df):
        data = []
        fs = []
        for row in df.iterrows():
            row = row[1]
            outp = self.request_data(row['path'], row['channel'], start=row['start'], stop=row['end'])
            if not outp[0]:
                raise AssertionError('[ERROR] Data haven\'t been read \n' + str(row) + ' \n Error Message: ' + str(outp))
            data += [outp[1]]
            fs += [outp[2]]
        return data, fs



    def request_metadata(self, path, passwd=''):
        client = self.client
        for p in self.ports:
            client.connect(f"tcp://{self.server_ip}:{p}")
        self.poll.register(client, zmq.POLLIN)
        try:
            client.send_json({'path': path,  'passwd': passwd, 'flag': 'basic_info'})
            socks = dict(self.poll.poll(self.response_wait * 1000))
            if socks.get(client) == zmq.POLLIN:
                md = client.recv_json(flags=0, )
                if md['error'] == 1:
                    return False, md['error_message']
                else:
                    return True, md['bi']
            else:
                client.setsockopt(zmq.LINGER, 0)
                client.close()
                self.poll.unregister(client)
                # rebind
                context = zmq.Context(1)
                self.client = context.socket(zmq.REQ)
                return False, f'Server response time elapsed {self.response_wait} s'

        except Exception as exc:
            exce = f'{exc}'
            return False, exce




# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from datetime import datetime, timedelta
from pandas._libs.tslibs.timestamps import Timestamp
from dateutil import tz
import pandas as pd
import numpy as np
from copy import deepcopy
from tqdm import tqdm


def _convert_to_timestamp(x):
    if isinstance(x, (datetime, Timestamp)):
        assert x.tzinfo, '[TIMEZONE ERROR] We allow operating with timezone-aware datatypes. This helps preventing inconsistency and errors.'
        return x.timestamp()
    if isinstance(x, (float, int)): return x
    raise TypeError('[TYPE ERROR]: input variable has to be of a type pandas Timestamp, datetime, float, or int. However ' + type(x) + ' recieved.')


def _convert_to_datetime_utc(x):
    if isinstance(x, (datetime, Timestamp)):
        assert x.tzinfo, '[TIMEZONE ERROR] We allow operating with timezone-aware datatypes. This helps preventing inconsistency and errors.'
        x = x.timestamp()
    if isinstance(x, (float, int)):
        utc = datetime.utcfromtimestamp(x)
        utc = utc.replace(tzinfo=tz.tzutc())

        return utc
    raise TypeError('[TYPE ERROR]: input variable has to be of a type pandas Timestamp, datetime, float, or int. However ' + type(x) + ' recieved.')


def _convert_to_pandas_timestamp_utc(x):
    if isinstance(x, (datetime, Timestamp)):
        assert x.tzinfo, '[TIMEZONE ERROR] We allow operating with timezone-aware datatypes. This helps preventing inconsistency and errors.'
        x = x.timestamp()
    if isinstance(x, (float, int)):
        utc = datetime.utcfromtimestamp(x)
        utc = utc.replace(tzinfo=tz.tzutc())
        utc = Timestamp(utc)
        return utc
    raise TypeError('[TYPE ERROR]: input variable has to be of a type pandas Timestamp, datetime, float, or int. However ' + type(x) + ' recieved.')


def _convert_to_utc(x):
    x = _convert_to_datetime_utc(x)
    return x


def _convert_to_local(x):
    x = _convert_to_datetime_utc(x)
    x = x.astimezone(tz.tzlocal())
    return x


def _convert_to_timezone(x, tzinfo):
    x = _convert_to_datetime_utc(x)
    x = x.astimezone(tzinfo)
    return x


def time_to_local(dfHyp):
    """
    Converts the time into the local timezone. Default by python and PC. Does not enter the timezone explicitely. Cannot be used for creating a hypnogram figure.
    """
    def convert(x, col_key):
        return _convert_to_local(x[col_key])

    dfHyp['start'] = dfHyp.apply(lambda x: convert(x, 'start'), axis=1)
    dfHyp['end'] = dfHyp.apply(lambda x: convert(x, 'end'), axis=1)
    return dfHyp


def time_to_utc(dfHyp):
    """
    Converts time to the UTC format.

    """

    def convert(x, col_key):
        return _convert_to_utc(x[col_key])

    dfHyp['start'] = dfHyp.apply(lambda x: convert(x, 'start'), axis=1)
    dfHyp['end'] = dfHyp.apply(lambda x: convert(x, 'end'), axis=1)
    return dfHyp


def time_to_timezone(dfHyp, tzinfo):
    """
    Converts the hypnogram into a timezone. The timezone has to be from a python library dateutil
    """
    def convert(x, col_key):
        return _convert_to_timezone(x[col_key], tzinfo)

    dfHyp['start'] = dfHyp.apply(lambda x: convert(x, 'start'), axis=1)
    dfHyp['end'] = dfHyp.apply(lambda x: convert(x, 'end'), axis=1)
    return dfHyp


def time_to_timestamp(dfHyp):
    """
    Converts the hypnogram time to timestamp.
    """
    def convert(x, col_key):
        return _convert_to_timestamp(x[col_key])

    dfHyp['start'] = dfHyp.apply(lambda x: convert(x, 'start'), axis=1)
    dfHyp['end'] = dfHyp.apply(lambda x: convert(x, 'end'), axis=1)
    return dfHyp


def create_duration(dfHyp):
    """
    Creates duration for each epoch within the hypnogram. (Faster on timestamp)
    """
    def duration(x):
        if type(x['start']) in (datetime, Timestamp):
            return _convert_to_timestamp(x['end']) - _convert_to_timestamp(x['start'])
        else:
            return x['end'] - x['start']
    dfHyp['duration'] = dfHyp.apply(lambda x: duration(x), axis=1)
    return dfHyp


def create_day_indexes(dfHyp, hour=12, tzinfo=tz.tzlocal):
    """
    Creates a day index for each epoch within the hypnogram, given the day-time hour supplied as input parameter.
    The format of start and end has to be an integer or a float in a form representing a timestamp, or a timezone aware datetime  or Timestamp object.
    If the start and end format do not include timezone, the local timezone will be used.
    """
    
    if not isinstance(dfHyp, pd.DataFrame):
        raise AssertionError('[INPUT ERROR]: Variable dfHyp must be of a type pandas.DataFrame.')

    if hour < 0 or hour > 23:
        raise ValueError(
            '[VALUE ERROR] - An input variable hour_cut indicating at which hour days are separated from each other must be on the range between 0 - 23. Pasted value: ',
            hour)

    timezone_counter = 0 # timezone_counter == dfHyp.__len__() if timeaware; == 0 if not timeaware; >0 & <dfHyp.__len__() if mismatch
    datetime_format = False
    tzinfo = None
    # check if the data is in
    for ridx, row in dfHyp.iterrows():
        if isinstance(row['start'], (Timestamp, datetime)) and isinstance(row['end'], (Timestamp, datetime)):
            if row['start'].tzinfo and (row['start'].tzinfo == row['start'].tzinfo == row['end'].tzinfo == row['end'].tzinfo):
                if ridx == 0:
                    tzinfo = row['start'].tzinfo
                    timezone_counter += 1
                elif row['start'].tzinfo == tzinfo:
                    timezone_counter += 1

    if timezone_counter > 0 and timezone_counter != dfHyp.__len__():
        raise ValueError('[VALUE ERROR] - Time zones in the start and end fields are inconsistent')


    dfHyp = dfHyp.sort_values('start').reset_index(drop=True)
    dfHyp['day'] = 0

    max_day = int(np.ceil((dfHyp.iloc[-1]['end'] - dfHyp.iloc[0]['start']).total_seconds() / (24*3600)))
    ref = dfHyp['start'][0].replace(hour=hour, minute=0, second=0, microsecond=0)

    for idx in range(max_day):
        dfHyp['day'][dfHyp['start'] >= ref] = idx + 1
        ref += timedelta(days=1)
    dfHyp['day'] -= dfHyp['day'].min()
    return dfHyp


def merge_annotations(df):
    """
    Merges epochs with the same annotation and end[i-1] == start[i]
    """
    new_df = pd.DataFrame()
    for idx, row in enumerate(df.iterrows()):
        appbl = True
        if idx > 0:
            if new_df.iloc[-1].annotation == row[1].annotation and new_df.iloc[-1].end == row[1].start:
                appbl = False

        if appbl == True:
            new_df = new_df.append(row[1], ignore_index=True)
        else:
            new_df.loc[new_df.__len__() - 1, 'end'] = row[1].end
        if type(new_df.loc[new_df.__len__() - 1, 'end']) in (datetime, Timestamp):
            new_df.loc[new_df.__len__() - 1, 'duration'] = (new_df.loc[new_df.__len__() - 1, 'end'] - new_df.loc[new_df.__len__() - 1, 'start']).seconds
        else:
            new_df.loc[new_df.__len__() - 1, 'duration'] = new_df.loc[new_df.__len__() - 1, 'end'] - new_df.loc[new_df.__len__() - 1, 'start']
    return new_df


def _tile_row(row, dur_threshold):
    outp = []
    start_time = row['start']
    end_time = row['end']
    curr_time = row['start']

    if type(row['end']) in (datetime, Timestamp):
        delta = timedelta(seconds=dur_threshold)
    else:
        delta = dur_threshold

    for idx in range(int(np.ceil(row['duration']/dur_threshold))):
        row_ = row.copy(deep=True)
        row_['start'] = curr_time
        curr_time += delta
        row_['end'] = curr_time
        if type(row_['end']) in (datetime, Timestamp):
            row_['duration'] = (row_['end'] - row_['start']).seconds
        else:
            row_['duration'] = row_['end'] - row_['start']

        outp += [row_]

    outp[-1]['end'] = end_time

    if type(outp[-1]['end']) in (datetime, Timestamp):
        outp[-1]['duration'] = (outp[-1]['end'] - outp[-1]['start']).seconds
    else:
        outp[-1]['duration'] = outp[-1]['end'] - outp[-1]['start']

    return outp


def tile_annotations(df, dur_threshold, verbose=False):
    """
    Tiles epochs to the max duration given by dur_threshold in seconds. Reverse to the 'merge annotations'.
    """
    if not isinstance(df, pd.DataFrame):
        raise AssertionError('[INPUT ERROR]: Variable dfAnnotations must be of type pandas.DataFrame.')

    if not isinstance(dur_threshold, (int, float)):
        raise AssertionError(
            '[INPUT ERROR]: dur_threshold must be float or int format giving the maximum duration of a single annotation. All anotations above this duration threshold will be tiled.')

    if np.isnan(dur_threshold) or np.isinf(dur_threshold) or dur_threshold <= 0:
        raise AssertionError('[INPUT ERROR]: dur_threshold must be a valid number bigger than 0, not nan and not inf')

    if (df['duration'] > dur_threshold).sum() > 1:
        outp = []
        if verbose:
            it = tqdm(list(df.iterrows()))
        else:
            it = list(df.iterrows())

        for row in it:
            row = row[1]
            outp += _tile_row(row, dur_threshold)
        return pd.DataFrame(outp).reset_index(drop=True)
    else:
        return df


def filter_by_duration(dfAnnotations, duration):
    """
    Keeps only epochs of the duration given by the input.
    """
    if not isinstance(dfAnnotations, pd.DataFrame):
        raise AssertionError('[INPUT ERROR]: Variable dfAnnotations must be of a type pandas.DataFrame.')

    if not isinstance(duration, (int, float)):
        raise AssertionError(
            '[INPUT ERROR]: duration must be float or int format giving the maximum duration of a single annotation. All anotations above this duration threshold will be tiled.')

    if np.isnan(duration) or np.isinf(duration) or duration <= 0:
        raise AssertionError('[INPUT ERROR]: duration must be a valid number bigger than 0, not nan and not inf')

    dfAnnotations = dfAnnotations.loc[dfAnnotations['duration'] == duration].reset_index(drop=True)
    return dfAnnotations


def filter_by_key(dfAnnotations, key, value):
    """
    Keeps only annotations given by the key and value within the pandas DataFrame
    """
    return dfAnnotations.loc[dfAnnotations[key] != value].reset_index(drop=True)







import torch
import numpy as np
from best.dbs.artifact_removal import configs_ArtifactEraser, models_ArtifactEraser
from best.signal import get_datarate, buffer

def remove_artifacts(x, fs, cuda=3):
    """

    :param x: numpy array with shape[N]
    :param fs: 250 or 500 Hz
    :param cuda: integer denoting id of Cuda to be used, alternatively 'cpu' is accepted as well
    :return:
    """

    model = models_ArtifactEraser[f'RCS_ArtifactEraser_{int(fs)}Hz_64']
    model = model.eval()

    if cuda == 'cpu':
        model = model.cpu()
    else:
        model = model.cuda(cuda)

    mn_x = np.nanmean(x)
    x -= mn_x
    nans = np.isnan(x)

    fs = int(fs)
    shape = x.shape[0]
    x = np.concatenate((np.zeros(1 * fs), x, np.zeros(5 * fs)))

    xb = buffer(x, fs, segm_size=2, overlap=1.0)
    mn = np.nanmean(xb, axis=1)
    mn[np.isnan(mn)] = 0

    std = np.nanstd(xb, axis=1)
    std[np.isnan(std)] = 1

    dr = get_datarate(xb)
    xb[np.isnan(xb)] = 0

    xb -= mn.reshape(-1, 1)  # zscore
    xb /= std.reshape(-1, 1)

    dr_pos = np.where(dr > 0.1)[0]  # only datarate > than 0.1 in 30 minute segment
    xb_proc = xb[dr_pos]
    xb[:] = np.NaN
 
    # idx_end = 1
    # if xb_proc.shape[0]-512 > 1:
    #     idx_end = xb_proc.shape[0]-512
    idx_end = x.shape[0]


    with torch.no_grad():
        for k in np.arange(0, idx_end, 512):
            tmp = xb_proc[k:k + 512]
            bs = tmp.shape[0]
            if bs > 0:
                if cuda == 'cpu':
                    tmp_rec, det = model(torch.tensor(tmp).float().view(bs, 1, -1).cpu())
                else:
                    tmp_rec, det = model(torch.tensor(tmp).float().view(bs, 1, -1).cuda(cuda))

                tmp_rec = tmp_rec.detach().view(tmp_rec.shape[0], tmp_rec.shape[2]).cpu().numpy()
                xb_proc[k:k + bs] = tmp_rec

    xb[dr_pos] = xb_proc
    xb *= std.reshape(-1, 1)
    xb += mn.reshape(-1, 1)  # denormalize

    xrec = xb[:, int(0.5 * fs):int(1.5 * fs)].flatten()[
           int(0.5 * fs):int(0.5 * fs) + shape]  # crop only valid semgments
    xrec[nans] = np.nan
    x = xrec
    x += mn_x
    return x





# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import numpy as np
from tqdm import tqdm
from copy import deepcopy, copy


import pandas as pd


from sklearn.feature_selection import RFECV, SelectFromModel
from sklearn.svm import SVR, LinearSVC
from sklearn import preprocessing
from sklearn.metrics import cohen_kappa_score, roc_curve

from scipy.stats import gaussian_kde
from scipy.signal import filtfilt
from scipy.signal.windows import gaussian
from scipy.optimize import differential_evolution
from scipy.linalg import norm


from best.hypnogram.utils import time_to_timestamp, time_to_utc, merge_annotations
from best.feature import augment_features, balance_classes
from best.signal import unify_sampling_frequency, get_datarate, buffer
from best.stats import kl_divergence_nonparametric
from best.vector import scale, translate, get_mutual_vectors
from best.modules import multivariate_normal_
from best.modules import ZScoreModule, PCAModule
from best.feature_extraction.FeatureExtractor import SleepSpectralFeatureExtractor
from best.feature_extraction.SpectralFeatures import mean_bands, mean_frequency, relative_bands
import scipy.signal as signal
from sklearn.model_selection import train_test_split
from sklearn.naive_bayes import GaussianNB

class KDEBayesianModel:
    __name__ = "KDEBayesianModel"
    def __init__(self, fbands=[[0.5, 5], # delta
                               [4, 9], # theta
                               [8, 14], # alpha
                               [11, 16], # spindle
                               [14, 20],
                               [20, 30]], segm_size=30, fs=200, bands_to_erase=[], filter_bands = True, nfft=12000,
                 window_smooth_n=3, window_std=1, cat_bias={'AWAKE': 1, 'N2': 1, 'N3': 1, 'REM': 1}, Selector2=True):

        self.fbands = fbands
        self.segm_size = segm_size
        self.fs = fs
        self.bands_to_erase = bands_to_erase
        self.filter_bands = filter_bands
        self.nfft=nfft
        self.filter_order = 100

        self.STATES = []
        self.KDE = []
        self.PipelineClustering = None
        self.FeatureSelector = None

        self.SELECTOR2 = Selector2

        self.FeatureExtractor_MeanBand = SleepSpectralFeatureExtractor(
            fs=self.fs,
            segm_size=self.segm_size,
            fbands=self.fbands,
            bands_to_erase=self.bands_to_erase,
            filter_bands=self.filter_bands,
            sperwelchseg=10,
            soverlapwelchseg=5,
            nfft=self.nfft,
            datarate=False
        )


        self.FeatureExtractor_MeanBand = SleepSpectralFeatureExtractor(
            fs=self.fs,
            segm_size=self.segm_size,
            fbands=self.fbands,
            bands_to_erase=self.bands_to_erase,
            filter_bands=self.filter_bands,
            sperwelchseg=10,
            soverlapwelchseg=5,
            nfft=self.nfft,
            datarate=False
        )




        self.FeatureExtractor_MeanBand._extraction_functions = \
            [
                mean_bands,
            ]



        self.FeatureExtractor = SleepSpectralFeatureExtractor(
            fs=self.fs,
            fbands=self.fbands,
            bands_to_erase=self.bands_to_erase,
            segm_size=self.segm_size,
            sperwelchseg=10,
            soverlapwelchseg=5,
            nfft=self.nfft,
            datarate=False
        )

        self.FeatureExtractor._extraction_functions = \
            [
                mean_frequency,
                #self.FeatureExtractor.MedFreq,
                relative_bands,
                #self.FeatureExtractor.normalized_entropy,
                #self.FeatureExtractor.normalized_entropy_bands
            ]


        self.WINDOW = gaussian(window_smooth_n, window_std)
        self.WINDOW = self.WINDOW / self.WINDOW.sum()
        self.CAT_BIAS = cat_bias
        self.feature_names = None


    def extract_features(self, signal, return_names=False):
        if signal.ndim > 1:
            raise AssertionError('[INPUT ERROR]: Input data has to be of a dimension size 1 - raw signal')
        if signal.shape[0] != self.fs * self.segm_size:
            print('[INPUT WARNING]: input data is not a defined size fs*segm_size ' + str(self.fs*self.segm_size) + '; Signal of a size ' + str(signal.shape[0]) + ' found instead. Extracted features might be inaccurate.')


        ## Mean band-derived features - delta/beta ratio etc
        mean_bands, feature_names = self.FeatureExtractor_MeanBand(signal)
        mean_bands = np.concatenate(mean_bands)

        functions = [np.divide]
        symbols = ['/']
        mean_band_derived_features, mean_band_derived_names = mean_bands, feature_names
        for idx in range(functions.__len__()):
            mean_band_derived_features, mean_band_derived_names = augment_features(

                mean_band_derived_features.reshape(1, -1), feature_indexes=np.arange(mean_band_derived_features.shape[0]), operation=functions[idx], mutual=True,  operation_str=symbols[idx], feature_names=mean_band_derived_names

            )


        mean_band_derived_names = mean_band_derived_names[feature_names.__len__():]
        mean_band_derived_features = mean_band_derived_features[0, feature_names.__len__():]
        #mean_band_derived_names = mean_band_derived_names.squeeze()

        #features = np.log10(np.append(other_features, mean_band_derived_features))
        #feature_names = feature_names + mean_band_derived_names
        features = np.log10(mean_band_derived_features)
        #feature_names = mean_band_derived_names

        ## other features
        other_features, feature_names_other = self.FeatureExtractor(signal)
        other_features = np.concatenate(other_features)
        features = np.append(other_features, features)
        feature_names = list(feature_names_other) + list(mean_band_derived_names)


        self.feature_names = feature_names
        if return_names:
            return features, feature_names
        return features

    def extract_features_bulk(self, list_of_signals, fsamp_list, return_names=False):
        data = list_of_signals
        data, fs = unify_sampling_frequency(data, sampling_frequency=fsamp_list, fs_new=self.fs)
        x = []
        for k in tqdm(range(data.__len__())):
            x += [self.extract_features(data[k])]
        if return_names:
            _, feature_names = self.extract_features(data[k], return_names=True)
            return np.array(x), feature_names
        return np.array(x), fs

    def fit(self, X, y):
        X, y = self._fit(X, y)
        self._fit_kde(X, y)

    def _fit(self, X, y):

        X = deepcopy(X)
        y = deepcopy(y)
        X_, y_ = balance_classes(X, y, std_factor=0.0)

        estimator = SVR(kernel="linear")
        self.SELECTOR = RFECV(estimator, step=5, verbose=True, min_features_to_select=4, n_jobs=10)
        self.PCA = PCAModule(var_threshold=0.98)
        self.ZScore = ZScoreModule(trainable=True, continuous_learning=False, multi_class=False)

        #self.UMAP = UMAP(n_neighbors=30, min_dist=1,
        #                 n_components=2)

        le = preprocessing.LabelEncoder()
        le.fit(y_)
        y__ = le.transform(y_)


        X_ = self.SELECTOR.fit_transform(X_, y__)
        X_ = self.PCA.fit_transform(X_)
        X_ = self.ZScore.fit_transform(X_, y)

        X = self.SELECTOR.transform(X)
        X = self.PCA.transform(X)
        X = self.ZScore.fit_transform(X, y)

        lsvc = LinearSVC(C=0.01, penalty="l1", dual=False).fit(X_, y_)
        if self.SELECTOR2:
            self.SELECTOR2 = SelectFromModel(lsvc, prefit=True, max_features=4)
            #X_ = self.SELECTOR2.transform(X_)
            X = self.SELECTOR2.transform(X)

        #X = self.UMAP.fit_transform(X)
        return X, y


    def _fit_kde(self, X, y):
        self.STATES = np.unique(y)
        self.KDE = []
        for state in self.STATES:
            X_ = X[y==state, :]
            kernel = gaussian_kde(X_.T)
            self.KDE.append(kernel)

    def _likelihood(self, X):
        scores = {}
        for idx, kde in enumerate(self.KDE):
            scores[self.STATES[idx]] = kde.pdf(X.T)
        scores = pd.DataFrame(scores)
        return scores

    def scores(self, X):
        X = self.transform(X)
        scores = self._likelihood(X)


        scores = scores.div(scores.sum(axis=1), axis=0)
        for key in scores.keys():
            scores[key] = filtfilt(self.WINDOW, 1, scores[key])

        for cat in self.CAT_BIAS.keys():
            if cat in scores.keys(): scores[cat] = scores[cat]*self.CAT_BIAS[cat]

        scores = scores.div(scores.sum(axis=1), axis=0)
        return scores

    def transform(self, X):
        X = self.SELECTOR.transform(X)
        X = self.PCA.transform(X)
        X = self.ZScore.transform(X)
        if self.SELECTOR2:
            X = self.SELECTOR2.transform(X)
        #X = self.UMAP.transform(X)
        return X

    def fit_transform(self, X, y):
        self.fit(X, y)
        return self.transform(X)

    def predict(self, X):
        return np.array(self.scores(X).idxmax(axis=1))

    def preprocess_signal(self, signal, fs, datarate_threshold=0.85):
        data = buffer(signal, fs, self.segm_size)
        start_time = np.array([k*self.segm_size for k in range(data.__len__())])
        end_time = start_time + self.segm_size
        datarate = np.array(get_datarate(data))

        data = data[datarate >= datarate_threshold]
        start_time = start_time[datarate >= datarate_threshold]
        end_time = end_time[datarate >= datarate_threshold]
        return list(data), start_time, end_time

    def predict_signal(self, signal, fs, datarate_threshold=0.85):
        data, start_time, end_time = self.preprocess_signal(signal, fs, datarate_threshold)
        x, fs = self.extract_features_bulk(data, [fs]*data.__len__())
        scores = self.scores(x)
        df = pd.DataFrame({'annotation': scores.idxmax(axis=1), 'start': start_time, 'end': start_time+30, 'duration':30})
        df = time_to_utc(df)
        df = merge_annotations(df)
        df = time_to_timestamp(df)
        df = df[['annotation', 'start', 'end', 'duration']]
        return df

    def predict_signal_scores(self, signal, fs, datarate_threshold=0.85):
        data, start_time, end_time = self.preprocess_signal(signal, fs, datarate_threshold)
        x, fs = self.extract_features_bulk(data, [fs]*data.__len__())
        scores = self.scores(x)
        return scores


class KDEBayesianCausalModel(KDEBayesianModel):
    __name__ = "KDEBayesianCausalModel"
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.MarkovFilter = None

    def fit(self, X, y):
        super().fit(X, y)

        scores = super().scores(X)
        self.MarkovFilter = SleepStageProbabilityMarkovChainFilter()
        self.MarkovFilter.fit(scores, y)


    def scores(self, X):
        scores = self._scores(X)
        state = 'AWAKE'
        scores = self.MarkovFilter.predict(scores, state)
        """
        ch_posts = []
        for k in range(scores.__len__()):
            p_likelihood_change = scores.iloc[k][self.MarkovFilter.STATES[self.MarkovFilter.STATES != state]].sum()
            p_prior_change = self.MarkovFilter.get_state_change_prior(state)
            p_post_change = (p_likelihood_change * p_prior_change) / ((p_likelihood_change * p_prior_change) + ((1-p_likelihood_change) * (1-p_prior_change)))
            ch_posts += [[p_post_change, state]]

            p_likelihood = scores.iloc[k][self.MarkovFilter.STATES[self.MarkovFilter.STATES != state]]
            p_prior = self.MarkovFilter.get_changing_state_priors(state)[self.MarkovFilter.STATES != state]
            p_post = p_likelihood*p_prior / sum(p_likelihood*p_prior)

            if p_post_change > 0.5:
                state = p_post.idxmax()
                scores.loc[k, state] = p_post_change
                scores.loc[k, self.MarkovFilter.STATES[self.MarkovFilter.STATES!=state]] = p_post*(1-p_post_change)
            else:
                scores.loc[k, state] = 1-p_post_change
                scores.loc[k, self.MarkovFilter.STATES[self.MarkovFilter.STATES!=state]] = p_post*p_post_change
            #yy[k] = state
        """
        return scores

    def _scores(self, X):
        return super().scores(X)


class MVGaussBayesianModel(KDEBayesianModel):
    __name__ = "MVGaussBayesianModel"

    def _fit_kde(self, X, y):
        self.STATES = np.unique(y)
        self.KDE = []
        for state in self.STATES:
            X_ = X[y==state, :]
            kernel = multivariate_normal_(X_.T)
            self.KDE.append(kernel)


class MVGaussBayesianCausalModel(MVGaussBayesianModel, KDEBayesianCausalModel):
    __name__ = "MVGaussBayesianModel"


class SleepStageProbabilityMarkovChainFilter:
    def __init__(self):
        self.STATES = np.array(['AWAKE', 'N1', 'N2', 'N3', 'REM'])
        self.removed_classes = []
        self.stability = np.ones(self.STATES.__len__())
        self._tmat_orig = np.array(
            [[0.961, 0.038, 0.001, 0.000, 0.000],
             [0.097, 0.215, 0.634, 0.000, 0.054],
             [0.020, 0.001, 0.846, 0.060, 0.073],
             [0.005, 0.001, 0.105, 0.880, 0.009],
             [0.017, 0.003, 0.061, 0.000, 0.918]]
        )

        #self._tmat_orig = np.array(
        #    [
        #        [0.802, 0.19,  0.007, 0.,    0.002],
        #        [0.11,  0.625, 0.24,  0.,    0.026],
        #        [0.018, 0.046, 0.89,  0.036, 0.01 ],
        #        [0.007, 0.006, 0.171, 0.815, 0.001],
        #        [0.017, 0.043, 0.012, 0.,    0.927]
        #    ])


        self.tmat = self._tmat_orig.copy()




        #np.fill_diagonal(self.tmat, self.tmat.diagonal()*stability)
        #diag = self.tmat.diagonal()
        #self.tmat = self.tmat / (self.tmat.sum(axis=1)-diag).reshape(-1, 1) * (1-diag).reshape(-1, 1)


        #diag = self.tmat.diagonal() * stability
        #self.tmat = self.tmat * (1-stability)
        #np.fill_diagonal(self.tmat, diag)
        self._get_vars()

    def fit(self, scores, y):
        self.reset_probabilities()
        self.tmat = self._tmat_orig.copy()
        classes = np.unique(y)
        #class_certainty = dict([[state, X[state][y==state].median()] for state in classes])

        for cl in self.STATES:
            if cl not in classes:
                self.remove_class(cl)

        self._tmat_orig = self.tmat.copy()

        #self.correct_certainty(class_certainty)

        stability = []
        for state in self.STATES:
            fpr, tpr, thresholds = roc_curve(np.array(y)==state, np.array(scores[state]))
            #fpr, tpr, thresholds = precision_recall_curve(np.array(y)==state, np.array(scores[state]))
            t = np.sqrt((1-fpr)**2 + tpr**2)
            #t = 2*(fpr*tpr) / (fpr+tpr)
            s = 1 - thresholds[t.argmax()]

            #fpr, tpr, thresholds = precision_recall_curve(np.array(y)==state, np.array(scores[state]))
            #t = np.sqrt(fpr**2 + tpr**2)
            #t = 2 * fpr * tpr / (fpr+tpr)
            #s = thresholds[t.argmax()]
            stability += [s]


            #precision_recall_curve
        #stability = self._optimize(scores, y)
        #stability = [0.51861717, 0.39611577, 1.09778543, 0.32874582]
        #stability = [0.12461134, 0.3359193,  0.18570281, 0.37578364]
        #stability = [0.10092481, 0.30177824, 0.21477364, 0.25693041]
        #stability = [0.06515025678829323, 0.3012126596534704, 0.18660074488436126, 0.2829450179411525]
        #stability = [0.75009757, 0.46948611, 0.53506914, 0.78013241]
        self.reset_probabilities()
        self.weight_probabilities(stability)
        self.stability = stability

    def fit_optimize(self, scores, y, Niter=200, popsize=10):
        self.fit(scores, y)

        stability = self._optimize(scores, y, Niter, popsize)
        self.reset_probabilities()
        self.weight_probabilities(stability)
        self.stability = stability


    def _get_vars(self):
        self.tmat = copy(self.tmat / self.tmat.sum(axis=1).reshape(-1, 1))
        self.change_prob = copy(1 - self.tmat.diagonal())
        self.prob_change_to = copy(self.tmat / (self.tmat.sum(axis=1) - self.tmat.diagonal()).reshape(-1, 1))
        self.prob_change_to[range(self.prob_change_to.shape[0]), range(self.prob_change_to.shape[1])] = 0

    def get_state_idx(self, state):
        return np.where(self.STATES == state)[0][0]

    def get_state_priors(self, state):
        idx = self.get_state_idx(state)
        priors = self.tmat[idx]
        return priors

    def get_state_change_prior(self, state):
        state_idx = self.get_state_idx(state)
        priors = self.get_state_priors(state)
        change_prior = 1 - priors[state_idx]
        return change_prior

    def get_prob_to_change(self, state, x_prob):
        idx = self.get_state_idx(state)
        prob_to_stay = x_prob[idx]
        return 1 - prob_to_stay

    def get_state_change_posterior(self, state, x_prob):
        prior = self.get_state_change_prior(state)
        prob = self.get_prob_to_change(state, x_prob)
        return prior * prob / (prior*prob + (1-prior)*(1-prob))


    def get_changing_state_priors(self, state):
        state_idx = copy(self.get_state_idx(state))
        priors = copy(self.get_state_priors(state))
        priors[state_idx] = 0
        priors = priors / priors.sum()
        return priors


    def get_changing_state_probabilities(self, state, x_prob):
        idx = self.get_state_idx(state)
        x_prob[idx] = 0
        return x_prob / x_prob.sum()


    def get_changing_state_posteriors(self, state, x_prob):
        priors = self.get_changing_state_priors(state)
        probs = self.get_changing_state_probabilities(state, x_prob)
        return probs*priors / np.sum(priors*probs)


    def correct_certainty(self, certainty: dict):
        for state in certainty.keys():
            idx = self.get_state_idx(state)
            cert = certainty[state]
            self.tmat[idx, idx] = cert * self.tmat[idx, idx]

        #diag = self.tmat.diagonal()
        #self.tmat = self.tmat / (self.tmat.sum(axis=1)-diag).reshape(-1, 1) * (1-diag).reshape(-1, 1)
        self._get_vars()


    def remove_class(self, class_name):
        idx = self.get_state_idx(class_name)
        self.tmat = np.delete(np.delete(self.tmat, idx, axis=0), idx, axis=1)

        diag = self.tmat.diagonal()
        self.tmat = self.tmat / (self.tmat.sum(axis=1)-diag).reshape(-1, 1) * (1-diag).reshape(-1, 1)
        np.fill_diagonal(self.tmat, diag)

        self.STATES = self.STATES[self.STATES != class_name]
        self._get_vars()
        self.removed_classes.append(class_name)


    def reset_probabilities(self):
        self.tmat = self._tmat_orig.copy()


    def weight_probabilities(self, stability: np.ndarray):
        #for idx in range(self.STATES.__len__()):
        #self.tmat[idx, idx] *= stability[idx]
        #diag = self.tmat.diagonal()
        #self.tmat = self.tmat / (self.tmat.sum(axis=1)-diag).reshape(-1, 1) * (1-diag).reshape(-1, 1)

        for idx in range(self.STATES.__len__()):
            self.tmat[idx, idx] = stability[idx]

        diag = self.tmat.diagonal()
        self.tmat = self.tmat / (self.tmat.sum(axis=1)-diag).reshape(-1, 1) * (1-diag).reshape(-1, 1)

        for idx in range(self.STATES.__len__()):
            self.tmat[idx, idx] = stability[idx]

        self._get_vars()


    def predict(self, scores, state='AWAKE'):
        ch_posts = []
        for k in range(scores.__len__()):
            p_likelihood_change = scores.iloc[k][self.STATES[self.STATES != state]].sum()
            p_prior_change = self.get_state_change_prior(state)
            p_post_change = (p_likelihood_change * p_prior_change) / ((p_likelihood_change * p_prior_change) + ((1-p_likelihood_change) * (1-p_prior_change)))
            ch_posts += [[p_post_change, state]]

            p_likelihood = scores.iloc[k][self.STATES[self.STATES != state]]
            p_prior = self.get_changing_state_priors(state)[self.STATES != state]
            p_post = p_likelihood*p_prior / sum(p_likelihood*p_prior)

            if p_post_change > 0.5:
                state = p_post.idxmax()
                scores.loc[k, state] = p_post_change
                scores.loc[k, self.STATES[self.STATES!=state]] = p_post*(1-p_post_change)
            else:
                scores.loc[k, state] = 1-p_post_change
                scores.loc[k, self.STATES[self.STATES!=state]] = p_post*p_post_change


        for k in np.where(np.isnan(scores))[0]:
            s = np.isnan(scores.iloc[k]).idxmax()
            scores.loc[k, s] = 1 - scores.iloc[k].sum()
        return scores


    def _optimize(self, scores, Y, Niter=100, popsize=20):
        print('Optimizing sleep stage stability using Differential Evolution')
        self.idx = -1
        self.best_kappa = -1
        self.best_stability = np.zeros(self.STATES.__len__())
        self.scores = scores
        self.Y = Y
        self.n_optimize=Niter

        def optimize_(args):
            self.idx += 1
            if self.idx < self.n_optimize:
                #args = [10**i for i in args]
                scores__ = deepcopy(self.scores)
                #self.reset_probabilities()
                self.weight_probabilities(args)
                sc_ = self.predict(scores__)
                y_ = np.array(sc_.idxmax(axis=1))
                kappa = cohen_kappa_score(self.Y, y_)
                sc = 1 - kappa
                #sc = np.sum([1-float(c) for c in list(get_classification_scores(self.Y, y_).values())])
                print(self.idx, kappa, args)

                if kappa > self.best_kappa:
                    self.best_kappa = kappa
                    self.best_stability = args
            else: sc = 0

            return sc




        bounds = [(0.2, 0.90)] * self.STATES.__len__()
        result = differential_evolution(optimize_, bounds, popsize=popsize, seed=1)
        result = self.best_stability

        del self.idx
        del self.best_kappa
        del self.best_stability
        del self.scores
        del self.Y

        return result


class KDEBayesianModelNC:
    __name__ = "KDEBayesianModel"
    def __init__(self, fbands=[[0.5, 5], # delta
                               [4, 9], # theta
                               [8, 14], # alpha
                               [11, 16], # spindle
                               [14, 20],
                               [20, 30]], segm_size=30, fs=200, bands_to_erase=[], filter_bands = True, filter_order=5001, nfft=12000,
                 window_smooth_n=3, window_std=1, cat_bias={'AWAKE': 1, 'N2': 1, 'N3': 1, 'REM': 1}, Selector2=True
                 ):

        self.fbands = fbands
        self.segm_size = segm_size
        self.fs = fs
        self.bands_to_erase = bands_to_erase
        self.filter_bands = filter_bands
        self.filter_order = filter_order
        self.nfft=nfft

        self.STATES = []
        self.KDE = []
        self.PipelineClustering = None
        self.FeatureSelector = None

        self.SELECTOR2 = Selector2

        self.FeatureExtractor_MeanBand = SleepSpectralFeatureExtractor(
            fs=self.fs,
            segm_size=self.segm_size,
            fbands=self.fbands,
            bands_to_erase=self.bands_to_erase,
            filter_bands=self.filter_bands,
            nfiltorder=self.filter_order,
            sperwelchseg=10,
            soverlapwelchseg=5,
            nfft=self.nfft,
            datarate=False
        )



        self.FeatureExtractor_MeanBand._extraction_functions = \
            [
                mean_bands,
            ]



        self.FeatureExtractor = SleepSpectralFeatureExtractor(
            fs=self.fs,
            fbands=self.fbands,
            bands_to_erase=self.bands_to_erase,
            segm_size=self.segm_size,
            sperwelchseg=10,
            soverlapwelchseg=5,
            nfft=self.nfft,
            datarate=False
        )
        self.FeatureExtractor._extraction_functions = \
            [
                mean_frequency,
                #self.FeatureExtractor.MedFreq,
                relative_bands,
                #self.FeatureExtractor.normalized_entropy,
                #self.FeatureExtractor.normalized_entropy_bands
            ]


        self.WINDOW = gaussian(window_smooth_n, window_std)
        self.WINDOW = self.WINDOW / self.WINDOW.sum()
        self.CAT_BIAS = cat_bias
        self.feature_names = None


    def extract_features(self, signal, return_names=False):
        if signal.ndim > 1:
            raise AssertionError('[INPUT ERROR]: Input data has to be of a dimension size 1 - raw signal')
        if signal.shape[0] != self.fs * self.segm_size:
            print('[INPUT WARNING]: input data is not a defined size fs*segm_size ' + str(self.fs*self.segm_size) + '; Signal of a size ' + str(signal.shape[0]) + ' found instead. Extracted features might be inaccurate.')


        ## Mean band-derived features - delta/beta ratio etc
        mean_bands, feature_names = self.FeatureExtractor_MeanBand(signal)
        mean_bands = np.concatenate(mean_bands)

        functions = [np.divide]
        symbols = ['/']
        mean_band_derived_features, mean_band_derived_names = mean_bands, feature_names
        for idx in range(functions.__len__()):
            mean_band_derived_features, mean_band_derived_names = augment_features(

                mean_band_derived_features.reshape(1, -1), feature_indexes=np.arange(mean_band_derived_features.shape[0]), operation=functions[idx], mutual=True,  operation_str=symbols[idx], feature_names=mean_band_derived_names

            )


        mean_band_derived_names = mean_band_derived_names[feature_names.__len__():]
        mean_band_derived_features = mean_band_derived_features[0, feature_names.__len__():]
        #mean_band_derived_names = mean_band_derived_names.squeeze()

        #features = np.log10(np.append(other_features, mean_band_derived_features))
        #feature_names = feature_names + mean_band_derived_names
        features = np.log10(mean_band_derived_features)
        #feature_names = mean_band_derived_names

        ## other features
        other_features, feature_names_other = self.FeatureExtractor(signal)
        other_features = np.concatenate(other_features)
        features = np.append(other_features, features)
        feature_names = list(feature_names_other) + list(mean_band_derived_names)


        self.feature_names = feature_names
        if return_names:
            return features, feature_names
        return features

    def extract_features_bulk(self, list_of_signals, fsamp_list, return_names=False):
        data = list_of_signals
        data, fs = unify_sampling_frequency(data, sampling_frequency=fsamp_list, fs_new=self.fs)
        x = []
        for k in tqdm(range(data.__len__())):
            x += [self.extract_features(data[k])]
        if return_names:
            _, feature_names = self.extract_features(data[k], return_names=True)
            return np.array(x), feature_names
        return np.array(x), fs

    def fit(self, X, y):
        X, y = self._fit(X, y)
        self._fit_kde(X, y)

    def _fit(self, X, y):

        X = deepcopy(X)
        y = deepcopy(y)
        X_, y_ = balance_classes(X, y, std_factor=0.0)

        estimator = SVR(kernel="linear")
        self.SELECTOR = RFECV(estimator, step=5, verbose=True, min_features_to_select=4, n_jobs=10)
        self.PCA = PCAModule(var_threshold=0.98)
        #self.ZScore = ZScoreModule(trainable=True, continuous_learning=False, multi_class=False)

        #self.UMAP = UMAP(n_neighbors=30, min_dist=1,
        #                 n_components=2)

        le = preprocessing.LabelEncoder()
        le.fit(y_)
        y__ = le.transform(y_)


        X_ = self.SELECTOR.fit_transform(X_, y__)
        X_ = self.PCA.fit_transform(X_)
        #X_ = self.ZScore.fit_transform(X_, y)

        X = self.SELECTOR.transform(X)
        X = self.PCA.transform(X)
       # X = self.ZScore.fit_transform(X, y)

        lsvc = LinearSVC(C=0.01, penalty="l1", dual=False).fit(X_, y_)
        if self.SELECTOR2:
            self.SELECTOR2 = SelectFromModel(lsvc, prefit=True, max_features=4)
            #X_ = self.SELECTOR2.transform(X_)
            X = self.SELECTOR2.transform(X)

        #X = self.UMAP.fit_transform(X)
        return X, y


    def _fit_kde(self, X, y):
        self.STATES = np.unique(y)
        self.KDE = []
        for state in self.STATES:
            X_ = X[y==state, :]
            kernel = gaussian_kde(X_.T)
            self.KDE.append(kernel)

    def _likelihood(self, X):
        scores = {}
        for idx, kde in enumerate(self.KDE):
            scores[self.STATES[idx]] = kde.pdf(X.T)
        scores = pd.DataFrame(scores)
        return scores

    def scores(self, X):
        X = self.transform(X)
        scores = self._likelihood(X)


        scores = scores.div(scores.sum(axis=1), axis=0)
        for key in scores.keys():
            scores[key] = filtfilt(self.WINDOW, 1, scores[key])

        for cat in self.CAT_BIAS.keys():
            if cat in scores.keys(): scores[cat] = scores[cat]*self.CAT_BIAS[cat]

        scores = scores.div(scores.sum(axis=1), axis=0)
        return scores

    def transform(self, X):
        X = self.SELECTOR.transform(X)
        X = self.PCA.transform(X)
        #X = self.ZScore.transform(X)
        if self.SELECTOR2:
            X = self.SELECTOR2.transform(X)
        #X = self.UMAP.transform(X)
        return X

    def fit_transform(self, X, y):
        self.fit(X, y)
        return self.transform(X)

    def predict(self, X):

        return np.array(self.scores(X).idxmax(axis=1))

    def preprocess_signal(self, signal, fs, datarate_threshold=0.85):
        data = buffer(signal, fs, self.segm_size)
        start_time = np.array([k*self.segm_size for k in range(data.__len__())])
        end_time = start_time + self.segm_size
        datarate = np.array(get_datarate(data))

        data = data[datarate >= datarate_threshold]
        start_time = start_time[datarate >= datarate_threshold]
        end_time = end_time[datarate >= datarate_threshold]
        return list(data), start_time, end_time

    def predict_signal(self, signal, fs, datarate_threshold=0.85):
        data, start_time, end_time = self.preprocess_signal(signal, fs, datarate_threshold)
        x, fs = self.extract_features_bulk(data, [fs]*data.__len__())
        scores = self.scores(x)
        df = pd.DataFrame({'annotation': scores.idxmax(axis=1), 'start': start_time, 'end': start_time+30, 'duration':30})
        df = time_to_utc(df)
        df = merge_annotations(df)
        df = time_to_timestamp(df)
        df = df[['annotation', 'start', 'end', 'duration']]
        return df

    def predict_signal_scores(self, signal, fs, datarate_threshold=0.85):
        data, start_time, end_time = self.preprocess_signal(signal, fs, datarate_threshold)
        x, fs = self.extract_features_bulk(data, [fs]*data.__len__())
        scores = self.scores(x)
        return scores


class Mapper:
    def __init__(self):
        self.TEMPLATE = None
        self.TEMPLATE_NAME = None
        self.CLASSES = None
        self.TEMPLATE_CLASS = None
        self.MAPS = dict()
        self.NDIM = None

        self.N = 1000

        self.SHIFT_RANGE = np.array([-5, 5])
        self.SCALE_RANGE = np.array([0.9, 1.25])
        self.ROTATE_RANGE = np.array([-25, 25])

        self.X_RANGE = []

        self.COREGISTRATION_CLASS_BAN = ['N1']

    def create_template(self, x, y=None, name='Template'):
        self.NDIM = x.shape[1]
        self._get_template(x)
        if not isinstance(y, type(None)):
            self._get_template_class(x, y)

        self.TEMPLATE_NAME = name

    def _get_template(self, x):
        self.X_RANGE = [x.mean() - 20*x.std(), x.mean() + 20*x.std()]
        self.TEMPLATE = self.get_probabilities(x)

    def _get_template_class(self, x, y):
        y = np.array(y)
        self.CLASSES = np.sort(np.unique(y))
        self.TEMPLATE_CLASS = dict([(cl, self.get_probabilities(x[y == cl])) for cl in self.CLASSES])

    def _cost(self, x):
        px = self.get_probabilities(x)
        return kl_divergence_nonparametric(self.TEMPLATE, px)

    def _cost_class(self, x, y, bias={}):
        cost = 0
        classes = list(np.unique(y))
        for cl in self.CLASSES:
            if not cl in self.COREGISTRATION_CLASS_BAN:
                if cl in classes:
                    px = self.get_probabilities(x[y == cl])
                    c_ = kl_divergence_nonparametric(self.TEMPLATE_CLASS[cl], px)
                    if cl in bias.keys():
                        c_ *= bias[cl]
                    cost += c_
        return cost

    def _cost_semi_supervised(self, x, y, bias={}):
        xs_ = x[y!='']
        ys_ = y[y!='']

        return self._cost_class(xs_, ys_, bias=bias) + self._cost(x)

    def fit_map(self, x, y=None, bias={'REM': 2}):
        if not isinstance(y, type(None)):
            x = deepcopy(x)
            y = deepcopy(x)
            x, y = balance_classes(x, y)

        np.random.seed(0)
        transforms = []
        kl = []
        for k in tqdm(range(self.N)):
            x_ = deepcopy(x)
            tr = self._get_random_transform()
            x_ = self._transform(x_, tr)
            if isinstance(y, type(None)):
                c = self._cost(x_)
            else:
                #c = self._cost_class(x_, y)
                c = self._cost_semi_supervised(x_, y, bias=bias)

            kl += [c]
            transforms += [tr]
        return np.array(kl), transforms


    def fit_genetic(self, x, y=None, popsize=15):
        def optimize(args):
            x_ = deepcopy(x)
            tr = {}

            tr['translate'] = np.array([args[i] for i in range(0, 3)])
            tr['scale'] = np.array([args[i] for i in range(3, 6)])
            x_ = self._transform(x_, tr)
            c = self._cost(x_)
            return c

        bounds = [tuple(self.SHIFT_RANGE)] * x.shape[1] + [tuple(self.SCALE_RANGE)] * x.shape[1]
        result = differential_evolution(optimize, bounds, popsize=popsize)

        tr = {}
        tr['translate'] = np.array([result.x[i] for i in range(0, 3)])
        tr['scale'] = np.array([result.x[i] for i in range(3, 6)])
        return tr, self._cost(self._transform(x, tr))

    def fit_genetic_likelihood(self, x, y=None, popsize=15, model=None):
        def optimize(args):
            x_ = deepcopy(x)
            tr = {}

            tr['translate'] = np.array([args[i] for i in range(0, 3)])
            tr['scale'] = np.array([args[i] for i in range(3, 6)])
            x_ = self._transform(x_, tr)
            c = model._likelihood(x_)
            c = 1000 - c.sum().sum()
            return c

        bounds = [tuple(self.SHIFT_RANGE)] * x.shape[1] + [tuple(self.SCALE_RANGE)] * x.shape[1]
        result = differential_evolution(optimize, bounds, popsize=popsize)

        tr = {}
        tr['translate'] = np.array([result.x[i] for i in range(0, 3)])
        tr['scale'] = np.array([result.x[i] for i in range(3, 6)])
        return tr, self._cost(self._transform(x, tr))


    #def fit_map(self, x, name, y=None, model=None):
        #tr, cost = self.fit_genetic(x)
    #    tr, cost = self.fit_genetic_likelihood(x, model=model)
    #    self.MAPS[name] = {'transformation': tr, 'cost': cost}

    def map(self, x, name, model=None):
        if not name in self.MAPS.keys():
            self.fit_map(x, name=name, model=model)

        x = self._transform(x, self.MAPS[name]['transformation'])
        return x


    def _transform(self, x, transform):
        x = deepcopy(x)
        x = scale(x, transform['scale'])
        x = translate(x, transform['translate'])
        # x = rotate(x, transform['rotate'])
        return x

    def _get_random_transform(self):
        transform = {
            'translate': np.random.rand(self.NDIM) * np.diff(self.SHIFT_RANGE) + self.SHIFT_RANGE[0],
            'scale': np.random.rand(self.NDIM) * np.diff(self.SCALE_RANGE) + self.SCALE_RANGE[0]
        }
        #if self.NDIM == 2:
        #    transform['rotate'] = np.random.rand() * np.diff(self.ROTATE_RANGE) + self.ROTATE_RANGE[0]
        #if self.NDIM == 3:
        #    transform['rotate'] = np.random.rand(3) * np.diff(self.ROTATE_RANGE) + self.ROTATE_RANGE[0]

        return transform

    def get_probabilities(self, x):
        rng = self.X_RANGE
        bins = 200
        ps = []
        for k in range(x.shape[1]):
            p, _ = np.histogram(x[:, k], bins, rng)
            ps += [p]

        ps = np.array(ps) / x.shape[0]
        ps[ps == 0] = 1e-9
        ps = ps / ps.sum(axis=1).reshape(-1, 1)
        return ps


class SleepStructureClassifier:
    def __init__(self, states=['WAKE', 'N1', 'N2', 'N3', 'REM']):
        self.STATES = states
        self.norml2 = None
        self.KDE = None
        self.N = None

    def fit(self, x, y):
        xmut, ymut = get_mutual_vectors(x, y)
        norml2 = norm(xmut, axis=1)
        unit_vect = xmut / norml2.reshape(-1, 1)
        xmut = np.concatenate((norml2.reshape(-1, 1), unit_vect), axis=1)
        for s1 in self.STATES:
            self.KDE[s1] = {}
            for s2 in self.STATES:
                s = s1 + '-' + s2
                self.KDE[s1][s2] = gaussian_kde(xmut[ymut==s, :].T)


    def scores(self, x):
        N = x.shape[1]
        xmut, ymut = get_mutual_vectors(x)

        probs = {}
        p = None
        for s1 in self.STATES:
            probs[s1] = {}
            for s2 in self.STATES:
                probs[s1][s2] = self.KDE[s1][s2].pdf(xmut.T)
            probs[s1] = pd.DataFrame(probs[s1])
            if isinstance(p, type(None)):
                p = probs[s1]
            else:
                p = p + probs[s1]

        return pd.DataFrame([p.iloc[k::N].sum() for k in range(N)])


class SleepClassifierWrapper:
    def __init__(self):
        self.MODEL = {}

    def train(self, X, df):
        datarate = get_datarate(X)
        df = df.loc[datarate > 0.85].reset_index(drop=True)
        X = X[datarate > 0.85]


        X = X[df.annotation != 'N1']
        df = df.loc[df.annotation != 'N1'].reset_index(drop=True)

        X = X[df.annotation != 'UNKNOWN']
        df = df.loc[df.annotation != 'UNKNOWN'].reset_index(drop=True)


        fs_downsample = 250
        fbands = [[0.5, 5],  # delta
                  [4, 9],  # theta
                  [8, 14],  # alpha
                  [11, 16],  # spindle
                  [14, 20],
                  [20, 30]]  # (beta3)

        bands_to_erase_2 = [[0, 0.5]] + [[k - 0.5, k + 0.5] for k in np.arange(2, 30, 2)]
        bands_to_erase_7 = [[0, 0.5], [6, 8], [13, 15], [20, 22], [27, 29]]


        #### Train 0 ####
        X0 = X[df.freq == 0]
        df0 = df.loc[df.freq == 0].reset_index(drop=True)
        model0 = KDEBayesianModel(fs=fs_downsample, fbands=fbands, bands_to_erase=[], Selector2=False)
        X0 = model0.extract_features_bulk(X0)
        Y0 = df0.annotation.to_numpy()
        model0.fit(X0, Y0)
        self.MODEL[0] = model0

        #### Train 2 ####
        X2 = X[(df.freq == 2) | (df.freq == 0)]
        df2 = df.loc[(df.freq == 2) | (df.freq == 0)].reset_index(drop=True)
        model2 = KDEBayesianModel(fs=fs_downsample, fbands=fbands, bands_to_erase=bands_to_erase_2, Selector2=False)
        X2 = model2.extract_features_bulk(X2)
        Y2 = df2.annotation.to_numpy()
        model2.fit(X2, Y2)
        self.MODEL[2] = model2

        #### Train 7 ####
        X7 = X[(df.freq == 7) | (df.freq == 0)]
        df7 = df.loc[(df.freq == 7) | (df.freq == 0)].reset_index(drop=True)
        model7 = KDEBayesianModel(fs=fs_downsample, fbands=fbands, bands_to_erase=bands_to_erase_7, Selector2=False)
        X7 = model7.extract_features_bulk(X7)
        Y7 = df7.annotation.to_numpy()
        model7.fit(X7, Y7)
        self.MODEL[7] = model7

        #### Train 72.5 ####
        X725 = X[(df.freq == 72.5)]
        df725 = df.loc[(df.freq == 72.5)].reset_index(drop=True)
        model725 = KDEBayesianModel(fs=fs_downsample, fbands=fbands, bands_to_erase=[], Selector2=False)
        X725 = model725.extract_features_bulk(X725)
        Y725 = df725.annotation.to_numpy()
        model725.fit(X725, Y725)
        self.MODEL[72.5] = model725

    def predict_signal(self, X, fs, stim_freq):
        assert (fs == 250 or fs == 500), 'Sampling frequency has to be 250 or 500 Hz!!!!'

        if fs == 500:
            b, a = signal.butter(6, 40, fs=fs, btype='low', analog=False)
            nans = np.isnan(X)
            X[nans] = np.nanmean(X)
            X = signal.filtfilt(b, a, X)
            X[nans] = np.nan
            X = X[::2]

        clf = self.MODEL[stim_freq]
        return clf.predict_signal(X, 250, 0.85)


class MultiChannelMVGaussBayesClassifier:
    __name__ = "KDEBayesianModel"

    def __init__(self, fbands=[[0.5, 5],  # delta
                               [4, 9],  # theta
                               [8, 14],  # alpha
                               [11, 16],  # spindle
                               [14, 20],
                               [20, 30]], segm_size=30, fs=200, bands_to_erase=[], filter_bands=True, nfft=12000,
                 window_smooth_n=3, window_std=1, cat_bias={'AWAKE': 1, 'N2': 1, 'N3': 1, 'REM': 1},
                 Selector2=True):

        self.fbands = fbands
        self.segm_size = segm_size
        self.fs = fs
        self.bands_to_erase = bands_to_erase
        self.filter_bands = filter_bands
        self.nfft = nfft
        self.filter_order = 100

        self.STATES = []
        self.KDE = []
        self.PipelineClustering = None
        self.FeatureSelector = None

        self.SELECTOR2 = Selector2

        self.FeatureExtractor_MeanBand = SleepSpectralFeatureExtractor(
            fs=self.fs,
            segm_size=self.segm_size,
            fbands=self.fbands,
            bands_to_erase=self.bands_to_erase,
            filter_bands=self.filter_bands,
            sperwelchseg=10,
            soverlapwelchseg=5,
            nfft=self.nfft,
            datarate=False
        )

        self.FeatureExtractor_MeanBand = SleepSpectralFeatureExtractor(
            fs=self.fs,
            segm_size=self.segm_size,
            fbands=self.fbands,
            bands_to_erase=self.bands_to_erase,
            filter_bands=self.filter_bands,
            sperwelchseg=10,
            soverlapwelchseg=5,
            nfft=self.nfft,
            datarate=False
        )

        self.FeatureExtractor_MeanBand._extraction_functions = \
            [
                mean_bands,
            ]

        self.FeatureExtractor = SleepSpectralFeatureExtractor(
            fs=self.fs,
            fbands=self.fbands,
            bands_to_erase=self.bands_to_erase,
            segm_size=self.segm_size,
            sperwelchseg=10,
            soverlapwelchseg=5,
            nfft=self.nfft,
            datarate=False
        )

        self.FeatureExtractor._extraction_functions = \
            [
                mean_frequency,
                relative_bands,
            ]

        self.WINDOW = gaussian(window_smooth_n, window_std)
        self.WINDOW = self.WINDOW / self.WINDOW.sum()
        self.CAT_BIAS = cat_bias
        self.feature_names = None

    def extract_features(self, signal, return_names=False):
        if signal.ndim > 1:
            raise AssertionError('[INPUT ERROR]: Input data has to be of a dimension size 1 - raw signal')
        if signal.shape[0] != self.fs * self.segm_size:
            print('[INPUT WARNING]: input data is not a defined size fs*segm_size ' + str(
                self.fs * self.segm_size) + '; Signal of a size ' + str(
                signal.shape[0]) + ' found instead. Extracted features might be inaccurate.')

        ## Mean band-derived features - delta/beta ratio etc
        mean_bands, feature_names = self.FeatureExtractor_MeanBand(signal)
        mean_bands = np.concatenate(mean_bands)

        functions = [np.divide]
        symbols = ['/']
        mean_band_derived_features, mean_band_derived_names = mean_bands, feature_names
        for idx in range(functions.__len__()):
            mean_band_derived_features, mean_band_derived_names = augment_features(

                mean_band_derived_features.reshape(1, -1),
                feature_indexes=np.arange(mean_band_derived_features.shape[0]), operation=functions[idx],
                mutual=True, operation_str=symbols[idx], feature_names=mean_band_derived_names

            )

        mean_band_derived_names = mean_band_derived_names[feature_names.__len__():]
        mean_band_derived_features = mean_band_derived_features[0, feature_names.__len__():]
        # mean_band_derived_names = mean_band_derived_names.squeeze()

        # features = np.log10(np.append(other_features, mean_band_derived_features))
        # feature_names = feature_names + mean_band_derived_names
        features = np.log10(mean_band_derived_features)
        # feature_names = mean_band_derived_names

        ## other features
        other_features, feature_names_other = self.FeatureExtractor(signal)
        other_features = np.concatenate(other_features)
        features = np.append(other_features, features)
        feature_names = list(feature_names_other) + list(mean_band_derived_names)

        self.feature_names = feature_names
        if return_names:
            return features, feature_names
        return features

    def extract_features_bulk(self, list_of_signals, fsamp_list, return_names=False):
        data = list_of_signals
        # data, fs = unify_sampling_frequency(data, sampling_frequency=fsamp_list, fs_new=self.fs)
        x = []
        for k in tqdm(range(data.__len__())):
            x += [self.extract_features(data[k])]
        if return_names:
            _, feature_names = self.extract_features(data[k], return_names=True)
            return np.array(x), feature_names
        return np.array(x), fsamp_list[0]

    def fit(self, X, y):
        self.classifier = GaussianNB()
        self.classifier.fit(X, y)

    def scores(self, X):
        scr = self.classifier.predict_proba(X)
        return pd.DataFrame([{c: scr[:, idx]} for idx, c in enumerate(self.classifier.classes_)])

    def fit_transform(self, X, y):
        self.fit(X, y)
        return self.transform(X)

    def predict(self, X):
        return np.array(self.scores(X).idxmax(axis=1))

    def preprocess_signal(self, signal, fs, datarate_threshold=0.85):
        data = buffer(signal, fs, self.segm_size)
        start_time = np.array([k * self.segm_size for k in range(data.__len__())])
        end_time = start_time + self.segm_size
        datarate = np.array(get_datarate(data))

        data = data[datarate >= datarate_threshold]
        start_time = start_time[datarate >= datarate_threshold]
        end_time = end_time[datarate >= datarate_threshold]
        return list(data), start_time, end_time

    def predict_signal(self, signal, fs, datarate_threshold=0.85):
        data, start_time, end_time = self.preprocess_signal(signal, fs, datarate_threshold)
        x, fs = self.extract_features_bulk(data, [fs] * data.__len__())
        scores = self.scores(x)
        df = pd.DataFrame(
            {'annotation': scores.idxmax(axis=1), 'start': start_time, 'end': start_time + 30, 'duration': 30})
        df = time_to_utc(df)
        df = merge_annotations(df)
        df = time_to_timestamp(df)
        df = df[['annotation', 'start', 'end', 'duration']]
        return df

    def predict_signal_scores(self, signal, fs, datarate_threshold=0.85):
        data, start_time, end_time = self.preprocess_signal(signal, fs, datarate_threshold)
        x, fs = self.extract_features_bulk(data, [fs] * data.__len__())
        scores = self.scores(x)
        return scores





